#// 
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#// 

from uvm import *


class vip_sequence(UVMSequence):

    def __init__(self, name):
        super().__init__(name)
        self.set_automatic_phase_objection(1)


#class vip_one_char_seq(vip_sequence):
    #
    #  `uvm_object_utils(vip_one_char_seq)
    #   
    #  def __init__(self, name="vip_one_char_seq")
    #    super().__init__(name)
    #  endfunction
    #  
    #async
    #  def body(self)
    #    `uvm_do(req)
    #  endtask
    #  
    #endclass


class vip_sentence_seq(vip_sequence):

    def __init__(self, name="vip_sentence_seq"):
        super().__init__(name)

    async def body(self):
        for i in range(128):
            uvm_do(self.req)

uvm_object_utils(vip_sentence_seq)


#class vip_idle_esc_seq(vip_sequence):
    #
    #  `uvm_object_utils(vip_idle_esc_seq)
    #
    #  def __init__(self, name="vip_idle_esc_seq")
    #    super().__init__(name)
    #  endfunction
    #  
    #async
    #  def body(self)
    #     repeat (128):
    #        `uvm_do_with(req, {chr inside {8'hE7, 8'h81}; })
    #     end
    #  endtask
    #  
from uvm.seq.uvm_sequence import *
from uvm.macros import *
    #endclass



# File header
# Header row2

import_var = "ABC"

# Also move this as a comment
# for 'comment_this_func.
#
# With extra empty line
def comment_this_func(aaa):
    return aaa * 2

## Move this as class comment docstring
## Add this to class
class ABC():

    def do_not_comment_this(self):
        pass

    # Nothing to do here
    test_var = ""

    # Is this indent?
    def __init__(self, name):  # EOL ignored
        self.name = name

    # Floating comments, not included

    # Returns something.
    # Accepts nothing.
    async def my_func(self):
        return "Something"

    # Block before var, must preserve
    # var indent
    my_arr = []

    #def no_comments(self, ok):
    #    ok = 2 * ok
    #    return ok

    # This function does X and Y.
    # Plus Z also.
    @classmethod
    def static_func(cls, some_args):
        # Preserve this as line comment, DONT TOUCH
        return 2 * some_args

    def has_existing_comment(self, a1, a2):
        """ Already docstring, Should not do anything
        """
        pass

    # Should convert into docstring
    def gets_docstring_too(self):
        raise Exception("Not done")
        # Do not move this to next function
    # Nor this one

    def no_comments_here(self):
        pass

    # Should convert to docstring
    def receives_docstring(self, a, b, c):
        # Started comment
        if a > b:
            # Need to do something
            c = b * a
        # Closing comments
        return (c * c)

    # Get comment also
    def short_func(self, k):
        pass


    def triple_dedent_at_end(self):
        if 1 == 0:
            if 2 == 3:
                pass
    # Misc class comments

    # Move this to docstring of 'must_comment' function
    def must_comment(self):
        pass

    # Misc trailing class comments, keep this

    # And also this

# KEEP1
# KEEP2
# KEEP3

# KEEP 4
# KEEP 5



#//
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import Timer

from uvm.comps.uvm_env import UVMEnv
from uvm.macros import *
from uvm.base.uvm_config_db import UVMConfigDb


class tb_env(UVMEnv):

    def __init__(self, name, parent=None):
        super().__init__(name, parent)


    def build_phase(self, phase):
        UVMConfigDb.set(None, "global_timer.*",    "timeout", 1000)
        UVMConfigDb.set(None, "global_timer.main", "timeout", 3000)
        UVMConfigDb.set(None, "global_timer.run",  "timeout", 0)


    
    async def reset_phase(self, phase):
        phase.raise_objection(self)
        await Timer(20, "NS")
        phase.drop_objection(self)


    
    async def configure_phase(self, phase):
        phase.raise_objection(self)
        await Timer(200, "NS")
        phase.drop_objection(self)


    
    async def main_phase(self, phase):
        phase.raise_objection(self)
        await Timer(1000, "NS")
        phase.drop_objection(self)


    
    async def shutdown_phase(self, phase):
        phase.raise_objection(self)
        await Timer(10, "NS")
        phase.drop_objection(self)


uvm_component_utils(tb_env)



# File to test uvm_events implementation with simulator

import cocotb
from cocotb.triggers import Timer, Join, Combine

from uvm_testlib import *
from uvm.base.uvm_event import UVMEvent


#@cocotb.coroutine
#def wait_events(events):
#    waiters = []
#    e = Event()
#    triggers = list(events)
#
#    # start a parallel task for each trigger
#    for t in triggers:
#        @cocotb.coroutine
#        def waiter(t=t):
#            try:
#                yield t.wait()
#            finally:
#                triggers.remove(t)
#                if not triggers:
#                    e.set()
#        waiters.append(cocotb.fork(waiter()))
#
#    # wait for the last waiter to complete
#    yield e.wait()
#    raise ReturnValue(self)

#@cocotb.coroutine
#def setters(events):
#    for e in events:
#        yield Timer(10)
#        e.set()

@cocotb.test(skip=False)
async def test_single_event(dut):
    e1 = UVMEvent('e1')
    #e2 = UVMEvent('e2')
    #yield wait_events([e1])

    
    async def notify_evt():
        await Timer(100, "NS")
        e1.trigger(123)
    cocotb.fork(notify_evt())
    await e1.wait_trigger()

    end_time = sv.realtime()
    data = e1.get_trigger_data()
    if end_time < 100 or data != 123:
        raise Exception('Test failed')


@cocotb.test(skip=False)
async def test_uvm_event(dut):
    evt1 = UVMEvent('MyEvent1')
    evt2 = UVMEvent('MyEvent2')
    n1 = cocotb.fork(notify_event(evt1, 10))
    n2 = cocotb.fork(notify_event(evt2, 20))
    await evt1.wait_trigger()
    d1 = evt1.get_trigger_data()
    await evt2.wait_trigger()
    d2 = evt2.get_trigger_data()
    if d1 != 10 or d2 != 20:
        raise Exception('Event data error. d1: {}, d2: {}'.format(d1, d2))


async def notify_event(evt, after=10):
    await Timer(after, "NS")
    evt.trigger(after)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm import (UVMComponent, sv, UVM_DEFAULT)
from uvm.macros import *


class C(UVMComponent):


    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.v = 0  # type: int
        self.s = 0  # type: int
        self.myaa = {}  # type: str
        self.my_conf_obj = None
        self.tag = ''


    def build_phase(self, phase):
        super().build_phase(phase)
        sv.display("%s: In Build: v = %0d  s = %0d", self.get_full_name(),
            self.v, self.s)


    #endclass


uvm_component_utils_begin(C)
uvm_field_int('v', UVM_DEFAULT)
uvm_field_int('s', UVM_DEFAULT)
uvm_field_object('my_conf_obj', UVM_DEFAULT)
uvm_field_string('tag', UVM_DEFAULT)
#uvm_field_aa_string_string('myaa', UVM_DEFAULT)
uvm_component_utils_end(C)




"""Fixer that changes line comments preceding funcdef into a docstring
inside that function.

Example:
    # Returns some stuff.
    # No args accepted.
    def my_func():

Output:
    def my_func():
        "" Returns some stuff.
        No args accepted.
        ""

"""

import re
from lib2to3 import pytree, fixer_base
from lib2to3.pgen2 import token
from lib2to3.fixer_util import (Assign, Attr, Name, is_tuple, is_list, syms,
    String, Newline)
from lib2to3.pygram import python_symbols

DEBUG = False

re_empty = re.compile(r'^\s*$')
re_dedent = re.compile(r'\n    #')
re_dedent2 = re.compile(r'\n#')
re_code_block = re.compile(r'^\s+#?\|')
re_hash_only = re.compile('^\s+#$')

rst_code_block = '.. code-block:: python'

def _debug(*args):
    if DEBUG is True:
        print(*args)


def lined(msg):
    return "|" + msg + "|"


class CommentStruct():

    def __init__(self, comm, no_funcs, node):
        self.comments = comm
        self.no_funcs = no_funcs.copy()
        self.node = node
        self.func_name = ""
        _debug("ComMStruct added no_funcs " + str(no_funcs) + ', node: '
            + str(node))

    def set_func(self, func):
        self.func_name = func


class FixFuncComments(fixer_base.BaseFix):

    explicit = False  # The user must ask for this fixers
    printed = False

    COMMA = pytree.Leaf(token.COMMA, ",")
    COLON = pytree.Leaf(token.COLON, ":")
    COMMENT = pytree.Leaf(token.COMMENT, '#')
    SEPS = (COMMA, COLON)

    re_comm = re.compile('#')
    re_hash_space = re.compile(r'\n# ')
    re_link = re.compile(r'~(\w+)~')
    re_link2 = re.compile(r'<(\w+)>')

    comments = []
    has_def = False
    has_name = False
    def_name = ""
    func_name_seen = {}

    class_stack = []

    func_args = {}
    func_return = {}
    func_raises = {}
    func_indent = {}


    def reset_state(self):
        self.has_def = False
        self.has_name = False


    def get_func_name(self, node):
        for cc in node.children:
            if cc.type == token.NAME:
                if cc.value != 'def':
                    return cc.value
        raise Exception("No name found for " + str(node))


    def should_add_to(self, node):
        if node.type == self.syms.funcdef:
            func_name = self.get_func_name(node)
            return func_name in self.func_name_seen


    def match(self, node):
        prefix = node.prefix
        if len(prefix) > 0:
            _debug("[match]: prefix is now |" + prefix + "|")

        if node.type == self.syms.classdef:
            self.class_stack.pop()
            while len(self.comments) > 0:
                self.restore_last_comment()
            self.reset_func_data()
            return False

        # Ignore nested classes
        if len(self.class_stack) > 1:
            return False

        # We detect function start here
        if node.type == token.NAME:
            if node.value == 'class':
                # TODO extract name properly
                self.class_stack.append('class')
                if len(self.class_stack) > 1:
                    print("WARN. Ignoring nested classes")
            if self.has_def is False and node.value == 'def':
                _debug("[match]: NAME with value |" + node.value + "|")
                self.has_def = True
            elif node.value == 'class':
                if len(self.comments) > 0:
                    self.restore_last_comment()
                return False
            elif self.has_def is True and self.has_name is False:
                self.has_name = True
                self.def_name = node.value
                self.func_indent[self.def_name] = 0
                _debug("[match]: Function " + node.value + " BEGIN")
                # Check here if we have indent/dedent comment available
                self.func_name_seen[node.value] = True

        # Extract Returns/Args/Raises for the docstring
        if node.type == self.syms.return_stmt:
            self.get_func_return(node)
        if node.type == self.syms.typedargslist:
            self.func_args[self.def_name] = self.get_func_args(node)
        if node.type == self.syms.raise_stmt:
            self.func_raises[self.def_name] = True


        # Discard previous comments if required
        if self.has_def is False and self.has_name is False:
            if node.type == self.syms.simple_stmt:
                if len(self.comments) > 0:
                    self.restore_last_comment()

        # Keep track of function indent
        if self.has_def is True and self.has_name is True:
            if node.type == token.INDENT:
                self.func_indent[self.def_name] += 1
            if node.type == token.DEDENT:
                self.func_indent[self.def_name] -= 1

        # Store comments from prefix, if accepted token
        if len(prefix) > 0 and self.re_comm.search(prefix):
            if self.check_node_type_for_comments(node) is False:
                return False

            # We're inside a function
            if self.has_def is True and self.has_name is True:
                if (self.func_indent[self.def_name] != 0 and
                        not re_dedent.search(prefix)):
                    _debug("[match]: Discard prefix, no indent match: " +
                        str(prefix))
                    return False
                if len(self.class_stack) == 0:
                    if (self.func_indent[self.def_name] != 0 and
                            not re_dedent2.search(prefix)):
                        _debug("[match]: Discard prefix (not in class), no indent match: " +
                            str(prefix))
                        return False

            keep_prefix, comments = self.split_prefix(prefix)
            if len(comments) > 0:
                comm_struct = CommentStruct(comments,
                        list(self.func_name_seen.keys()), node)
                comm_struct.set_func(self.def_name)
                self.comments.append(comm_struct)  # TODO strip comment signs #
                node.prefix = keep_prefix
                _debug("[match] Stored comments: |" + comments + "|")
                _debug("[match] Preserved prefix: |" + keep_prefix + "|")
            else:
                _debug("[match] Ignored empty comments: |" + comments + "|")
            return False
        elif self.should_add_to(node):  # TODO match function type
            _debug("[match] Found funcdef |" + str(node) + "|")
            return True
        return False


    def transform(self, func_node, results):
        #new = func_node.clone()
        new = func_node
        _debug('transform(): Started for function ' + lined(self.def_name))
        indent = ""

        if len(self.comments) == 0:
            self.reset_state()
            return new

        for child in new.children:
            if child.type == self.syms.suite:
                suite_node = child
                _debug("Found suite! Indent is |" + indent + "|")
                found_ind = False

                for cc2 in suite_node.children:
                    # Need to skip first NEWLINE token
                    if found_ind is False:
                        if cc2.type == token.INDENT:
                            indent = cc2.value  # Preserve correct indent
                            _debug("In suite.children! Updated indent to |" + indent + "|")
                            found_ind = True
                    else:
                        func_name = self.get_func_name(func_node)
                        comm_struct = self.comments[0]
                        curr_comments = comm_struct.comments
                        if func_name not in comm_struct.no_funcs:

                            if self.re_hash_space.search(curr_comments):
                                indented_text = curr_comments.replace('# ', indent)
                            else:
                                indented_text = curr_comments.replace('\n    ', '\n' + indent)

                            indented_text = self.format_docstring(func_name,
                                    indented_text, indent)
                            comments = (indent + '""" ' + indented_text +
                                '\n' + indent + '"""')
                            suite_children = [String(comments), Newline()]
                            suite_node.insert_child(1,
                                    pytree.Node(syms.simple_stmt, suite_children))
                            self.comments.pop(0)
                        else:
                            # If comment was found after current func, we want
                            # to keep it for the next function, else restore
                            if comm_struct.func_name != func_name:
                                self.restore_last_comment()
                            self.reset_state()
                            return func_node
                        break
                _debug("Breaking outer for-loop after suite")
                break
        _debug("New node would be |" + str(new) + "|")
        self.reset_state()
        return new


    def split_prefix(self, prefix):
        """ Splits the prefix string in two parts, one will be preserved as
        original prefix, and 2nd part will be the new comments """
        keep_prefix = ""
        comments = ""
        lines = prefix.split("\n")
        _debug("split_prefix() lines are " + str(lines))
        before, after = self.split_empty(lines[0:-1])

        last_line = lines[-1]
        before.append('')
        before.append(last_line)

        keep_prefix = "\n".join(before)
        comments = self.join_comments(after)
        return keep_prefix, comments


    def split_empty(self, lines):
        """ If there are 2 sections of comments, need to split that into 2 or
        more parts. FInd the last empty line, then use that line as split
        point.
        """
        before = []
        after = []
        last_empty = -1
        for i, line in enumerate(lines):
            if re_empty.match(line):
                _debug("split_empty EMPTY LINE @" + str(i))
                last_empty = i

        if last_empty != -1:
            before = lines[0:last_empty]
            after = lines[last_empty:]
        else:
            after = lines
        return before, after


    def join_comments(self, lines):
        return "\n".join(lines)


    def check_node_type_for_comments(self, node):
        return (node.type == token.NAME or node.type == token.INDENT
            or node.type == token.DEDENT or node.type == token.AT)


    def restore_last_comment(self):
        comm_struct = self.comments.pop(0)
        node = comm_struct.node
        split = node.prefix.split("\n")
        _debug("restore() split before merge " + str(split))
        new_prefix = ("\n").join(split[0:-1])
        new_prefix += comm_struct.comments
        new_prefix += "\n" + split[-1]
        node.prefix = new_prefix
        _debug("Recreated node.prefix as " +
            lined(comm_struct.node.prefix))


    def get_func_args(self, node):
        res = []
        for child in node.children:
            if child.type == token.NAME and child.value != 'self':
                res.append(child.value)
        return res


    def get_func_return(self, node):
        res = []
        if self.def_name in self.func_return:
            res = self.func_return[self.def_name]
        else:
            self.func_return[self.def_name] = res
        res.append('Returns')


    def format_docstring(self, func_name, text, indent):
        indented_text = text.replace('//', '')
        indented_text = indented_text.replace('# ', '')
        split = indented_text.split("\n")

        new_split = []
        in_code_block = False
        for i, line in enumerate(split):
            split[i] = self.re_link.sub(r'`\1`', line)
            split[i] = self.re_link2.sub(r'`\1`', split[i])

            if re_code_block.search(split[i]):
                split[i] = split[i].replace('#|', '')
                if in_code_block is False:
                    in_code_block = True
                    new_split.append(indent + rst_code_block)
                    new_split.append('')
            else:
                in_code_block = False

            if len(split[i]) < len(indent):
                split[i] = indent + split[i]
            split[i] = re_hash_only.sub('', split[i])
            new_split.append(split[i])

        if func_name in self.func_args:
            args = self.func_args[func_name]
            new_split.append(indent + "Args:")
            for arg in args:
                if arg not in ['None', 'True', 'False']:
                    new_split.append(indent + "    " + arg + ": ")
        if func_name in self.func_return:
            new_split.append(indent + "Returns:")
        if func_name in self.func_raises:
            new_split.append(indent + "Raises:")
        return "\n".join(new_split)


    def reset_func_data(self):
        """ Resets all collected function data """
        self.comments = []
        self.reset_state()
        self.def_name = ""
        self.func_name_seen = {}
        self.func_args = {}
        self.func_return = {}
        self.func_raises = {}
        self.func_indent = {}



# type: ignore
#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//-------------.----------------------------------------------------------------
"""
Title: uvm_*_imp ports

The following defines the TLM implementation (imp) classes.

CLASS: uvm_*_imp #(T,IMP)

Unidirectional implementation (imp) port classes. An imp port provides access
to an implementation of the associated interface to all connected ~ports~ and
~exports~. Each imp port instance ~must~ be connected to the component instance
that implements the associated interface, typically the imp port's parent.
All other connections-- e.g. to other ports and exports are prohibited.

The asterisk in ~uvm_*_imp~ may be any of the following::

    blocking_put
    nonblocking_put
    put

    blocking_get
    nonblocking_get
    get

    blocking_peek
    nonblocking_peek
    peek

    blocking_get_peek
    nonblocking_get_peek
    get_peek

The interface methods are implemented in a component of type ~IMP~, a handle
to which is passed in a constructor argument.  The imp port delegates all
interface calls to this component.

Function: __init__

Creates a new unidirectional imp port with the given ~name~ and ~parent~.
The ~parent~ must implement the interface associated with this port.
Its type must be the type specified in the imp's type-parameter, ~IMP~::

    def __init__(self, name, parent):

"""

from .uvm_tlm_imps import (UVM_BLOCKING_GET_IMP, UVM_BLOCKING_GET_PEEK_IMP, UVM_BLOCKING_PEEK_IMP,
                           UVM_BLOCKING_PUT_IMP, UVM_BLOCKING_TRANSPORT_IMP, UVM_GET_IMP,
                           UVM_GET_PEEK_IMP, UVM_IMP_COMMON, UVM_MS_IMP_COMMON,
                           UVM_NONBLOCKING_GET_IMP, UVM_NONBLOCKING_GET_PEEK_IMP,
                           UVM_NONBLOCKING_PEEK_IMP, UVM_NONBLOCKING_PUT_IMP,
                           UVM_NONBLOCKING_TRANSPORT_IMP, UVM_PEEK_IMP, UVM_PUT_IMP)
from ..base.uvm_port_base import UVMPortBase
from ..macros.uvm_tlm_defines import (UVM_TLM_BLOCKING_GET_MASK, UVM_TLM_BLOCKING_GET_PEEK_MASK,
                                      UVM_TLM_BLOCKING_MASTER_MASK, UVM_TLM_BLOCKING_PEEK_MASK,
                                      UVM_TLM_BLOCKING_PUT_MASK, UVM_TLM_BLOCKING_SLAVE_MASK,
                                      UVM_TLM_BLOCKING_TRANSPORT_MASK, UVM_TLM_GET_MASK,
                                      UVM_TLM_GET_PEEK_MASK, UVM_TLM_MASTER_MASK,
                                      UVM_TLM_NONBLOCKING_GET_MASK,
                                      UVM_TLM_NONBLOCKING_GET_PEEK_MASK,
                                      UVM_TLM_NONBLOCKING_MASTER_MASK,
                                      UVM_TLM_NONBLOCKING_PEEK_MASK, UVM_TLM_NONBLOCKING_PUT_MASK,
                                      UVM_TLM_NONBLOCKING_SLAVE_MASK,
                                      UVM_TLM_NONBLOCKING_TRANSPORT_MASK, UVM_TLM_PEEK_MASK,
                                      UVM_TLM_PUT_MASK, UVM_TLM_SLAVE_MASK, UVM_TLM_TRANSPORT_MASK)


#class uvm_blocking_put_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_BLOCKING_PUT_MASK,"uvm_blocking_put_imp",IMP)
#  `UVM_BLOCKING_PUT_IMP (m_imp, T, t)
#endclass
class UVMBlockingPutImp():
    pass
UVMBlockingPutImp = UVM_IMP_COMMON(UVMBlockingPutImp, UVM_TLM_BLOCKING_PUT_MASK, "uvm_blocking_put_imp")
UVM_BLOCKING_PUT_IMP('m_imp', UVMBlockingPutImp)

#class uvm_nonblocking_put_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_NONBLOCKING_PUT_MASK,"uvm_nonblocking_put_imp",IMP)
#  `UVM_NONBLOCKING_PUT_IMP (m_imp, T, t)
#endclass
class UVMNonBlockingPutImp():
    pass
UVMNonBlockingPutImp = UVM_IMP_COMMON(UVMNonBlockingPutImp, UVM_TLM_NONBLOCKING_PUT_MASK, "uvm_nonblocking_put_imp")
UVM_NONBLOCKING_PUT_IMP('m_imp', UVMNonBlockingPutImp)

#class uvm_put_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_PUT_MASK,"uvm_put_imp",IMP)
#  `UVM_PUT_IMP (m_imp, T, t)
#endclass
class UVMPutImp():
    pass
UVMPutImp = UVM_IMP_COMMON(UVMPutImp, UVM_TLM_PUT_MASK,"uvm_put_imp")
UVM_PUT_IMP('m_imp', UVMPutImp)

#class uvm_blocking_get_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_BLOCKING_GET_MASK,"uvm_blocking_get_imp",IMP)
#  `UVM_BLOCKING_GET_IMP (m_imp, T, t)
#endclass
class UVMBlockingGetImp():
    pass
UVMBlockingGetImp = UVM_IMP_COMMON(UVMBlockingGetImp, UVM_TLM_BLOCKING_GET_MASK,"uvm_blocking_get_imp")
UVM_BLOCKING_GET_IMP('m_imp', UVMBlockingGetImp)

#class uvm_nonblocking_get_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_NONBLOCKING_GET_MASK,"uvm_nonblocking_get_imp",IMP)
#  `UVM_NONBLOCKING_GET_IMP (m_imp, T, t)
#endclass
class UVMNonBlockingGetImp():
    pass
UVMNonBlockingGetImp = UVM_IMP_COMMON(UVMNonBlockingGetImp, UVM_TLM_NONBLOCKING_GET_MASK,"uvm_nonblocking_get_imp")
UVM_NONBLOCKING_GET_IMP('m_imp', UVMNonBlockingGetImp)

#class uvm_get_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_GET_MASK,"uvm_get_imp",IMP)
#  `UVM_GET_IMP (m_imp, T, t)
#endclass
class UVMGetImp():
    pass
UVMGetImp = UVM_IMP_COMMON(UVMGetImp, UVM_TLM_GET_MASK,"uvm_get_imp")
UVM_GET_IMP('m_imp', UVMGetImp)

#class uvm_blocking_peek_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_BLOCKING_PEEK_MASK,"uvm_blocking_peek_imp",IMP)
#  `UVM_BLOCKING_PEEK_IMP (m_imp, T, t)
class UVMBlockingPeekImp():
    pass
UVMBlockingPeekImp = UVM_IMP_COMMON(UVMBlockingPeekImp, UVM_TLM_BLOCKING_PEEK_MASK,"uvm_blocking_peek_imp")
UVM_BLOCKING_PEEK_IMP('m_imp', UVMBlockingPeekImp)

#class uvm_nonblocking_peek_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_NONBLOCKING_PEEK_MASK,"uvm_nonblocking_peek_imp",IMP)
#  `UVM_NONBLOCKING_PEEK_IMP (m_imp, T, t)
class UVMNonBlockingPeekImp():
    pass
UVMNonBlockingPeekImp = UVM_IMP_COMMON(UVMNonBlockingPeekImp, UVM_TLM_NONBLOCKING_PEEK_MASK, "uvm_nonblocking_peek_imp")
UVM_NONBLOCKING_PEEK_IMP('m_imp', UVMNonBlockingPeekImp)

#class uvm_peek_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_PEEK_MASK,"uvm_peek_imp",IMP)
#  `UVM_PEEK_IMP (m_imp, T, t)
#endclass
class UVMPeekImp():
    pass
UVMPeekImp = UVM_IMP_COMMON(UVMPeekImp, UVM_TLM_PEEK_MASK,"uvm_peek_imp")
UVM_PEEK_IMP ('m_imp', UVMPeekImp)

#class uvm_blocking_get_peek_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_BLOCKING_GET_PEEK_MASK,"uvm_blocking_get_peek_imp",IMP)
#  `UVM_BLOCKING_GET_PEEK_IMP (m_imp, T, t)
class UVMBlockingGetPeekImp():
    pass
UVMBlockingGetPeekImp = UVM_IMP_COMMON(UVMBlockingGetPeekImp, UVM_TLM_BLOCKING_GET_PEEK_MASK, "uvm_blocking_get_peek_imp")
UVM_BLOCKING_GET_PEEK_IMP ('m_imp', UVMBlockingGetPeekImp)

#class uvm_nonblocking_get_peek_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_NONBLOCKING_GET_PEEK_MASK,"uvm_nonblocking_get_peek_imp",IMP)
#  `UVM_NONBLOCKING_GET_PEEK_IMP (m_imp, T, t)
class UVMNonBlockingGetPeekImp():
    pass
UVMNonBlockingGetPeekImp = UVM_IMP_COMMON(UVMNonBlockingGetPeekImp,UVM_TLM_NONBLOCKING_GET_PEEK_MASK,"uvm_nonblocking_get_peek_imp")
UVM_NONBLOCKING_GET_PEEK_IMP ('m_imp', UVMNonBlockingGetPeekImp)

#class uvm_get_peek_imp #(type T=int, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_IMP_COMMON(`UVM_TLM_GET_PEEK_MASK,"uvm_get_peek_imp",IMP)
#  `UVM_GET_PEEK_IMP (m_imp, T, t)
class UVMGetPeekImp():
    pass
UVMGetPeekImp = UVM_IMP_COMMON(UVMGetPeekImp, UVM_TLM_GET_PEEK_MASK,"uvm_get_peek_imp")
UVM_GET_PEEK_IMP('m_imp', UVMGetPeekImp)

#//------------------------------------------------------------------------------
#//
#// CLASS: uvm_*_imp #(REQ, RSP, IMP, REQ_IMP, RSP_IMP)
#//
#// Bidirectional implementation (imp) port classes--An imp port provides access
#// to an implementation of the associated interface to all connected ~ports~ and
#// ~exports~. Each imp port instance ~must~ be connected to the component instance
#// that implements the associated interface, typically the imp port's parent.
#// All other connections-- e.g. to other ports and exports-- are prohibited.
#//
#// The interface represented by the asterisk is any of the following
#//
#//|  blocking_transport
#//|  nonblocking_transport
#//|  transport
#//|
#//|  blocking_master
#//|  nonblocking_master
#//|  master
#//|
#//|  blocking_slave
#//|  nonblocking_slave
#//|  slave
#//
#// Type parameters
#//
#// REQ  - Request transaction type
#//
#// RSP  - Response transaction type
#//
#// IMP  - Component type that implements the interface methods, typically the
#//        the parent of this imp port.
#//
#// REQ_IMP - Component type that implements the request side of the
#//           interface. Defaults to IMP. For master and slave imps only.
#//
#// RSP_IMP - Component type that implements the response side of the
#//           interface. Defaults to IMP. For master and slave imps only.
#//
#// The interface methods are implemented in a component of type ~IMP~, a handle
#// to which is passed in a constructor argument.  The imp port delegates all
#// interface calls to this component.
#//
#// The master and slave imps have two modes of operation.
#//
#// - A single component of type IMP implements the entire interface for
#//   both requests and responses.
#//
#// - Two sibling components of type REQ_IMP and RSP_IMP implement the request
#//   and response interfaces, respectively.  In this case, the IMP parent
#//   instantiates this imp port ~and~ the REQ_IMP and RSP_IMP components.
#//
#// The second mode is needed when a component instantiates more than one imp
#// port, as in the <uvm_tlm_req_rsp_channel #(REQ,RSP)> channel.
#//
#//------------------------------------------------------------------------------


#// Function: new
#//
#// Creates a new bidirectional imp port with the given ~name~ and ~parent~.
#// The ~parent~, whose type is specified by ~IMP~ type parameter,
#// must implement the interface associated with this port.
#//
#// Transport imp constructor
#//
#//|  function new(string name, IMP imp)
#//
#// Master and slave imp constructor
#//
#// The optional ~req_imp~ and ~rsp_imp~ arguments, available to master and
#// slave imp ports, allow the requests and responses to be handled by different
#// subcomponents. If they are specified, they must point to the underlying
#// component that implements the request and response methods, respectively.
#//
#//|  function new(string name, IMP imp,
#//|                            REQ_IMP req_imp=imp, RSP_IMP rsp_imp=imp)
#

#class uvm_blocking_master_imp #(type REQ=int, type RSP=REQ, type IMP=int,
#                                type REQ_IMP=IMP, type RSP_IMP=IMP)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  typedef IMP     this_imp_type;
#  typedef REQ_IMP this_req_type;
#  typedef RSP_IMP this_rsp_type;
#  `UVM_MS_IMP_COMMON(`UVM_TLM_BLOCKING_MASTER_MASK,"uvm_blocking_master_imp")
#  `UVM_BLOCKING_PUT_IMP (m_req_imp, REQ, t)
#  `UVM_BLOCKING_GET_PEEK_IMP (m_rsp_imp, RSP, t)
#endclass
class UVMBlockingMasterImp():  #(type REQ=int, type RSP=REQ, type IMP=int,
    pass
UVMBlockingMasterImp =  UVM_MS_IMP_COMMON(UVMBlockingMasterImp, UVM_TLM_BLOCKING_MASTER_MASK,"uvm_blocking_master_imp")
UVM_BLOCKING_PUT_IMP('m_req_imp', UVMBlockingMasterImp)
UVM_BLOCKING_GET_PEEK_IMP('m_rsp_imp', UVMBlockingMasterImp)


#class uvm_nonblocking_master_imp #(type REQ=int, type RSP=REQ, type IMP=int,
#                                   type REQ_IMP=IMP, type RSP_IMP=IMP)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  typedef IMP     this_imp_type;
#  typedef REQ_IMP this_req_type;
#  typedef RSP_IMP this_rsp_type;
#  `UVM_MS_IMP_COMMON(`UVM_TLM_NONBLOCKING_MASTER_MASK,"uvm_nonblocking_master_imp")
#  `UVM_NONBLOCKING_PUT_IMP (m_req_imp, REQ, t)
#  `UVM_NONBLOCKING_GET_PEEK_IMP (m_rsp_imp, RSP, t)
#endclass
class UVMNonBlockingMasterImp():  # (type REQ=int, type RSP=REQ, type IMP=int,
    pass
UVMNonBlockingMasterImp =  UVM_MS_IMP_COMMON(UVMNonBlockingMasterImp, UVM_TLM_NONBLOCKING_MASTER_MASK,"uvm_nonblocking_master_imp")
UVM_NONBLOCKING_PUT_IMP ('m_req_imp', UVMNonBlockingMasterImp)
UVM_NONBLOCKING_GET_PEEK_IMP ('m_rsp_imp', UVMNonBlockingMasterImp)


#class uvm_master_imp #(type REQ=int, type RSP=REQ, type IMP=int,
#                       type REQ_IMP=IMP, type RSP_IMP=IMP)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  typedef IMP     this_imp_type;
#  typedef REQ_IMP this_req_type;
#  typedef RSP_IMP this_rsp_type;
#  `UVM_MS_IMP_COMMON(`UVM_TLM_MASTER_MASK,"uvm_master_imp")
#  `UVM_PUT_IMP (m_req_imp, REQ, t)
#  `UVM_GET_PEEK_IMP (m_rsp_imp, RSP, t)
#endclass
class UVMMasterImp:
    pass
UVMMasterImp = UVM_MS_IMP_COMMON(UVMMasterImp, UVM_TLM_MASTER_MASK, "uvm_master_imp")
UVM_PUT_IMP('m_req_imp', UVMMasterImp)
UVM_GET_PEEK_IMP('m_rsp_imp', UVMMasterImp)


#class uvm_blocking_slave_imp #(type REQ=int, type RSP=REQ, type IMP=int,
#                               type REQ_IMP=IMP, type RSP_IMP=IMP)
#  extends uvm_port_base #(uvm_tlm_if_base #(RSP, REQ));
#  typedef IMP     this_imp_type;
#  typedef REQ_IMP this_req_type;
#  typedef RSP_IMP this_rsp_type;
#  `UVM_MS_IMP_COMMON(`UVM_TLM_BLOCKING_SLAVE_MASK,"uvm_blocking_slave_imp")
#  `UVM_BLOCKING_PUT_IMP (m_rsp_imp, RSP, t)
#  `UVM_BLOCKING_GET_PEEK_IMP (m_req_imp, REQ, t)
#endclass
class UVMBlockingSlaveImp():
    pass
UVMBlockingSlaveImp = UVM_MS_IMP_COMMON(UVMBlockingSlaveImp, 
    UVM_TLM_BLOCKING_SLAVE_MASK,"uvm_blocking_slave_imp")
UVM_BLOCKING_PUT_IMP('m_rsp_imp', UVMBlockingSlaveImp)
UVM_BLOCKING_GET_PEEK_IMP ('m_req_imp', UVMBlockingSlaveImp)

#class uvm_nonblocking_slave_imp #(type REQ=int, type RSP=REQ, type IMP=int,
#                                  type REQ_IMP=IMP, type RSP_IMP=IMP)
#  extends uvm_port_base #(uvm_tlm_if_base #(RSP, REQ));
#  typedef IMP     this_imp_type;
#  typedef REQ_IMP this_req_type;
#  typedef RSP_IMP this_rsp_type;
#  `UVM_MS_IMP_COMMON(`UVM_TLM_NONBLOCKING_SLAVE_MASK,"uvm_nonblocking_slave_imp")
#  `UVM_NONBLOCKING_PUT_IMP (m_rsp_imp, RSP, t)
#  `UVM_NONBLOCKING_GET_PEEK_IMP (m_req_imp, REQ, t)
#endclass
class UVMNonBlockingSlaveImp():   # (type REQ=int, type RSP=REQ, type IMP=int,
    pass
UVMNonBlockingSlaveImp = UVM_MS_IMP_COMMON(UVMNonBlockingSlaveImp,
    UVM_TLM_NONBLOCKING_SLAVE_MASK,"uvm_nonblocking_slave_imp")
UVM_NONBLOCKING_PUT_IMP('m_rsp_imp', UVMNonBlockingSlaveImp)
UVM_NONBLOCKING_GET_PEEK_IMP('m_req_imp', UVMNonBlockingSlaveImp)

#class uvm_slave_imp #(type REQ=int, type RSP=REQ, type IMP=int,
#                      type REQ_IMP=IMP, type RSP_IMP=IMP)
#  extends uvm_port_base #(uvm_tlm_if_base #(RSP, REQ));
#  typedef IMP     this_imp_type;
#  typedef REQ_IMP this_req_type;
#  typedef RSP_IMP this_rsp_type;
#  `UVM_MS_IMP_COMMON(`UVM_TLM_SLAVE_MASK,"uvm_slave_imp")
#  `UVM_PUT_IMP (m_rsp_imp, RSP, t)
#  `UVM_GET_PEEK_IMP (m_req_imp, REQ, t)
#endclass
class UVMSlaveImp:
    pass
UVMSlaveImp = UVM_MS_IMP_COMMON(UVMSlaveImp, UVM_TLM_SLAVE_MASK, "uvm_slave_imp")
UVM_PUT_IMP ('m_rsp_imp', UVMSlaveImp)
UVM_GET_PEEK_IMP('m_req_imp', UVMSlaveImp)


#class uvm_blocking_transport_imp #(type REQ=int, type RSP=REQ, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_IMP_COMMON(`UVM_TLM_BLOCKING_TRANSPORT_MASK,"uvm_blocking_transport_imp",IMP)
#  `UVM_BLOCKING_TRANSPORT_IMP (m_imp, REQ, RSP, req, rsp)
#endclass
class UVMBlockingTransportImp():  # (type REQ=int, type RSP=REQ, type IMP=int)
    pass
UVMBlockingTransportImp = UVM_IMP_COMMON(UVMBlockingTransportImp,
    UVM_TLM_BLOCKING_TRANSPORT_MASK,"uvm_blocking_transport_imp")
UVM_BLOCKING_TRANSPORT_IMP('m_imp', UVMBlockingTransportImp)

#class uvm_nonblocking_transport_imp #(type REQ=int, type RSP=REQ, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_IMP_COMMON(`UVM_TLM_NONBLOCKING_TRANSPORT_MASK,"uvm_nonblocking_transport_imp",IMP)
#  `UVM_NONBLOCKING_TRANSPORT_IMP (m_imp, REQ, RSP, req, rsp)
#endclass
class UVMNonBlockingTransportImp():  # (type REQ=int, type RSP=REQ, type IMP=int)
    pass
UVMNonBlockingTransportImp = UVM_IMP_COMMON(UVMNonBlockingTransportImp,
    UVM_TLM_NONBLOCKING_TRANSPORT_MASK,"uvm_nonblocking_transport_imp")
UVM_NONBLOCKING_TRANSPORT_IMP ('m_imp', UVMNonBlockingTransportImp)

#class uvm_transport_imp #(type REQ=int, type RSP=REQ, type IMP=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_IMP_COMMON(`UVM_TLM_TRANSPORT_MASK,"uvm_transport_imp",IMP)
#  `UVM_BLOCKING_TRANSPORT_IMP (m_imp, REQ, RSP, req, rsp)
#  `UVM_NONBLOCKING_TRANSPORT_IMP (m_imp, REQ, RSP, req, rsp)
#endclass
class UVMTransportImp:
    pass
UVMTransportImp = UVM_IMP_COMMON(UVMTransportImp, UVM_TLM_TRANSPORT_MASK, "uvm_transport_imp")
UVM_BLOCKING_TRANSPORT_IMP('m_imp', UVMTransportImp)
UVM_NONBLOCKING_TRANSPORT_IMP('m_imp', UVMTransportImp)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer

from uvm.base.sv import sv, semaphore
from uvm.base import UVMComponent, UVM_HIGH, UVM_MEDIUM, UVM_INFO
from uvm.tlm1 import UVMBlockingPutImp, UVMGetPort
from uvm.macros import uvm_component_utils, uvm_info


class consumer(UVMComponent):
    #
    #  uvm_blocking_put_imp #(T,consumer #(T)) in
    #  uvm_get_port #(T) out

    def __init__(self, name, parent=None):
        UVMComponent.__init__(self, name, parent)
        self.input = UVMBlockingPutImp("in",self)
        self.out = UVMGetPort("out",self,0)
        self.count = 0
        self.lock = semaphore(1)

    
    async def run_phase(self, phase):
        while self.out.size():
            print("consumer run_phase in while-loop")
            p = []
            await self.out.get(p)
            await self.put(p[0])

    
    async def put(self, p):
        print("consumer put() called with count " + str(self.count))
        await self.lock.get()
        self.count += 1
        self.accept_tr(p)
        await Timer(10, "NS")
        #    #10
        self.begin_tr(p)
        #    #30; 
        await Timer(10, "NS")
        self.end_tr(p)
        uvm_info("consumer", sv.sformatf("Received %0s local_count=%0d",
            p.get_name(),self.count), UVM_MEDIUM)
        if self.uvm_report_enabled(UVM_HIGH,UVM_INFO,""):
             p.print_obj()
        self.lock.put()
    #endclass

uvm_component_utils(consumer)
#    `uvm_field_int(count,UVM_ALL_ON + UVM_READONLY + UVM_DEC)
#  `uvm_component_utils_end




import cocotb
from cocotb.triggers import Timer
from uvm import sv


async def proc1(args):
    await Timer(10, "NS")
    args['ok1'] = True


async def proc2(args):
    await Timer(15, "NS")
    args['ok2'] = True
    await Timer(5, "NS")


@cocotb.test()
async def test_fork_join_any(dut):
    state = {'ok1': False, 'ok2': False}
    fork1 = cocotb.fork(proc1(state))
    fork2 = cocotb.fork(proc2(state))
    await sv.fork_join_any([fork1, fork2])
    if state['ok1'] is False:
        raise Exception('ok1 must be True')
    if state['ok2'] is True:
        raise Exception('ok2 must be False due to fork_join_any')
    await Timer(20, "NS")
    if state['ok2'] is False:
        raise Exception('ok2 must be True after 20ns')


@cocotb.test()
async def test_fork_join(dut):
    state = {'ok1': False, 'ok2': False}
    fork1 = cocotb.fork(proc1(state))
    fork2 = cocotb.fork(proc2(state))
    await sv.fork_join([fork1, fork2])
    if state['ok1'] is False:
        raise Exception('ok1 must be True')
    if state['ok2'] is False:
        raise Exception('ok2 must be True due to fork_join')

#@cocotb.test()
#async def test_fork_join_none(dut):
#    fork1 = cocotb.fork(proc1(state))
#    fork2 = cocotb.fork(proc2(state))
#    forks = sv.fork_join_none([proc1(state), proc2(state)])


@cocotb.test()
async def test_forks_kill(dut):
    state = {'ok1': False, 'ok2': False}
    fork1 = cocotb.fork(proc1(state))
    fork2 = cocotb.fork(proc2(state))
    await sv.fork_join_any([fork1, fork2])
    fork2.kill()
    await Timer(20, "NS")
    if state['ok1'] is False:
        raise Exception('ok1 must be True')
    if state['ok2'] is True:
        raise Exception('ok2 must be False due to kill()')



#// 
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2013      NVIDIA Corporation
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
#
#`ifndef UVM_DAP_SVH
# `define UVM_DAP_SVH
#
#// Set/Get DAPS
# `include "dap/uvm_set_get_dap_base.svh"
# `include "dap/uvm_simple_lock_dap.svh"
# `include "dap/uvm_get_to_lock_dap.svh"
# `include "dap/uvm_set_before_get_dap.svh"
#
#`endif // UVM_DAP_SVH
#



#------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010-2011 Synopsys, Inc.
#   Copyright 2013      NVIDIA Corporation
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------

from typing import List

import cocotb
from cocotb.triggers import Timer, Event

from .sv import sv, uvm_split_string
from .uvm_component import UVMComponent
from .uvm_version import (uvm_cdn_copyright, uvm_cy_copyright, uvm_mgc_copyright, uvm_nv_copyright,
                          uvm_revision_string, uvm_snps_copyright, uvm_tpoikela_copyright)
from .uvm_cmdline_processor import UVMCmdlineProcessor
from .uvm_globals import (uvm_is_match, uvm_report_error, uvm_report_warning,
    uvm_zero_delay)
from .uvm_object_globals import (UVM_DEBUG, UVM_ERROR, UVM_FULL, UVM_HIGH,
    UVM_LOW, UVM_MEDIUM, UVM_NONE)
from .uvm_phase import UVMPhase
from .uvm_debug import uvm_debug
from .uvm_objection import UVMObjection
from .uvm_report_server import UVMReportServer
from .uvm_domain import end_of_elaboration_ph
from .uvm_common_phases import UVMEndOfElaborationPhase
from ..macros import uvm_info, uvm_fatal, UVM_DEFAULT_TIMEOUT
from ..uvm_macros import UVM_STRING_QUEUE_STREAMING_PACK
from .uvm_config_db import UVMConfigDb

MULTI_TESTS = ("Multiple ({}) +UVM_TESTNAME arguments provided on the command"
        + "line. '{}' will be used.  Provided list: {}.")

WARN_MSG1 = ("Multiple (%0d) +UVM_VERBOSITY arguments provided on the command"
        + " line.  '%s' will be used.  Provided list: %s.")

NON_STD_VERB = "Non-standard verbosity value, using provided '%0d'."

LINE = "----------------------------------------------------------------"

UVM_OBJECT_DO_NOT_NEED_CONSTRUCTOR = 0

MSG_MULTTIMOUT = ("Multiple (%0d) +UVM_TIMEOUT arguments provided on the command line. "
    + "'%s' will be used.  Provided list: %s.")

MSG_INVLCMDARGS = ("Invalid number of arguments found on the command line for setting "
    + "'+uvm_set_verbosity=%s'.  Setting ignored.")

WARN_MAX_QUIT_COUNT = ("Multiple (%0d) +UVM_MAX_QUIT_COUNT arguments provided "
    + "on the command line.  '%s' will be used.  Provided list: %s.")


class UVMRoot(UVMComponent):
    """
    The `UVMRoot` class serves as the implicit top-level and phase controller for
    all UVM components. Users do not directly instantiate `UVMRoot`. The UVM
    automatically creates a single instance of <uvm_root> that users can
    access via the global (uvm_pkg-scope) variable, ~uvm_top~.

    (see uvm_ref_root.gif)

    The ~uvm_top~ instance of ~uvm_root~ plays several key roles in the UVM.

    Implicit top-level - The ~uvm_top~ serves as an implicit top-level component.
    Any component whose parent is specified as ~None~ becomes a child of ~uvm_top~.
    Thus, all UVM components in simulation are descendants of ~uvm_top~.

    Phase control - ~uvm_top~ manages the phasing for all components.

    Search - Use ~uvm_top~ to search for components based on their
    hierarchical name. See `UVMRoot.find` and `UVMRoot.find_all`.

    Report configuration - Use ~uvm_top~ to globally configure
    report verbosity, log files, and actions. For example,
    ~uvm_top.set_report_verbosity_level_hier(UVM_FULL)~ would set
    full verbosity for all components in simulation.

    Global reporter - Because ~uvm_top~ is globally accessible (in uvm_pkg
    scope), UVM's reporting mechanism is accessible from anywhere
    outside ~uvm_component~, such as in modules and sequences.
    See <uvm_report_error>, <self.uvm_report_warning>, and other global
    methods.


    The ~uvm_top~ instance checks during the end_of_elaboration phase if any errors have
    been generated so far. If errors are found a UVM_FATAL error is being generated as result
    so that the simulation will not continue to the start_of_simulation_phase.
    """

    raise_exception_on_die = True
    m_relnotes_done = False

    @classmethod
    def get(cls) -> 'UVMRoot':
        """
        Static accessor for `UVMRoot`.

        The static accessor is provided as a convenience wrapper
        around retrieving the root via the `UVMCoreService.get_root`
        method::

            # Using the uvm_coreservice_t:
            uvm_coreservice_t cs
            uvm_root r
            cs = uvm_coreservice_t::get()
            r = cs.get_root()

            # Not using the uvm_coreservice_t:
            uvm_root r
            r = uvm_root::get()

        Returns:
            UVMRoot: Handle to the UVMRoot singleton.

        """
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        return cs.get_root()

    m_inst = None

    def __init__(self):
        UVMComponent.__init__(self, "__top__", None)
        #self.m_children = {}
        self.clp = UVMCmdlineProcessor.get_inst()
        #  // Variable: finish_on_completion
        #  //
        #  // If set, then run_test will call $finish after all phases are executed.
        self.finish_on_completion = True
        self.m_phase_all_done = False
        self.m_phase_all_done_event = Event('phase_all_done_event')
        # If set, then the entire testbench topology is printed just after completion
        # of the end_of_elaboration phase.
        self.enable_print_topology = False

        self.m_rh.set_name("reporter")
        self.report_header()

        #  // This sets up the global verbosity. Other command line args may
        #  // change individual component verbosity.
        self.m_check_verbosity()

        #  // Variable: top_levels
        #  //
        #  // This variable is a list of all of the top level components in UVM. It
        #  // includes the uvm_test_top component that is created by <run_test> as
        #  // well as any other top level components that have been instantiated
        #  // anywhere in the hierarchy.
        self.top_levels: List[UVMComponent] = []
        #  // Variable- phase_timeout
        #  //
        #  // Specifies the timeout for the run phase. Default is `UVM_DEFAULT_TIMEOUT
        self.phase_timeout = UVM_DEFAULT_TIMEOUT

        # Handle to the DUT (added for uvm-python)
        self._dut = None


    m_called_get_common_domain = False

    @classmethod
    def m_uvm_get_root(cls):
        """        internal function not to be used
           get the initialized singleton instance of uvm_root
        Returns:
        """
        from .uvm_domain import UVMDomain
        if UVMRoot.m_inst is None:
            UVMRoot.m_inst = UVMRoot()
        # tpoikela: Added flag to avoid infinite recursion
        if cls.m_called_get_common_domain is False:
            cls.m_called_get_common_domain = True
            UVMDomain.get_common_domain()
            UVMRoot.m_inst.m_domain = UVMDomain.get_uvm_domain()
            cls.m_called_get_common_domain = False
        return UVMRoot.m_inst


    #@cocotb.coroutine
    #def run_phase(self):
    # TODO at checks that sim_time == 0

    def get_type_name(self):
        return "uvm_root"

    def final_phase(self, phase):
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        tr_db = cs.get_default_tr_database()
        if tr_db.is_open():
            tr_db.close_db()


    @classmethod
    def get_dut(cls):
        return cls.get()._dut
    #----------------------------------------------------------------------------
    # Group: Simulation Control
    #----------------------------------------------------------------------------


    async def run_test(self, test_name="", dut=None):
        """
        Phases all components through all registered phases. If the optional
        test_name argument is provided, or if a command-line plusarg,
        +UVM_TESTNAME=TEST_NAME, is found, then the specified component is created
        just prior to phasing. The test may contain new verification components or
        the entire testbench, in which case the test and testbench can be chosen from
        the command line without forcing recompilation. If the global (package)
        variable, finish_on_completion, is set, then $finish is called after
        phasing completes.

        Args:
            test_name (str): Name of the test to run.
            dut: Handle to the DUT.
        """
        uvm_debug(self, 'run_test', 'Called with testname |' + test_name + '|')
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        factory = cs.get_factory()
        testname_plusarg = False
        test_names = []
        msg = ""
        uvm_test_top = None  # uvm_component
        self._dut = dut
        ##process phase_runner_proc; // store thread forked below for final cleanup

        # from .uvm_scheduler import UVMScheduler
        # cocotb.fork(UVMScheduler.get().run_event_loop())

        # Set up the process that decouples the thread that drops objections from
        # the process that processes drop/all_dropped objections. Thus, if the
        # original calling thread (the "dropper") gets killed, it does not affect
        # drain-time and propagation of the drop up the hierarchy.
        # Needs to be done in run_test since it needs to be in an
        # initial block to fork a process.
        uvm_debug(self, 'run_test', 'Calling m_init_objections')
        await UVMObjection.m_init_objections()

        # Retrieve the test names provided on the command line.  Command line
        # overrides the argument.
        test_name_count = self.clp.get_arg_values("+UVM_TESTNAME=", test_names)

        uvm_debug(self, 'run_test', 'Found testnames from cmdline: ' +
                str(test_names))
        # If at least one, use first in queue.
        if test_name_count > 0:
            test_name = test_names[0]
            uvm_debug(self, 'run_test', 'Found test name %s' % (test_name))
            testname_plusarg = True

        # If multiple, provided the warning giving the number, which one will be
        # used and the complete list.
        if test_name_count > 1:
            test_list = ""
            sep = ""
            for i in range(0, len(test_names)):
                if i != 0:
                    sep = ", "
                test_list = test_list + sep + test_names[i]
            self.uvm_report_warning("MULTTST", MULTI_TESTS.format(
                test_name_count, test_name, test_list), UVM_NONE)

        # if test now defined, create it using common factory
        uvm_debug(self, 'run_test', 'Running now test ' + test_name)
        if test_name != "":
            if "uvm_test_top" in self.m_children:
                uvm_fatal("TTINST",
                    "An uvm_test_top already exists via a previous call to run_test")
            #0; // forces shutdown because $finish is forked
            await uvm_zero_delay()
            uvm_debug(self, 'run_test', "factory.create in UVMRoot testname " + test_name)

            uvm_test_top = factory.create_component_by_name(test_name,
                "", "uvm_test_top", None)

            if uvm_test_top is None:
                if testname_plusarg:
                    msg = "command line +UVM_TESTNAME=" + test_name
                else:
                    msg = "call to run_test(" + test_name + ")"
                uvm_fatal("INVTST", "Requested test from " + msg + " not found.")
        await uvm_zero_delay()

        if len(self.m_children) == 0:
            uvm_fatal("NOCOMP", ("No components instantiated. You must either instantiate"
                   + " at least one component before calling run_test or use"
                   + " run_test to do so. To run a test using run_test,"
                   + " use +UVM_TESTNAME or supply the test name in"
                   + " the argument to run_test(). Exiting simulation."))
            return

        self.running_test_msg(test_name, uvm_test_top)

        # phase runner, isolated from calling process
        #fork begin
        # spawn the phase runner task
        #phase_runner_proc = process::self()
        uvm_debug(self, 'run_test', 'Forking now phases')
        cocotb.fork(UVMPhase.m_run_phases())
        uvm_debug(self, 'run_test', 'After phase-fork executing')
        #end
        #join_none
        await uvm_zero_delay()
        ##0; // let the phase runner start

        uvm_debug(self, 'run_test', 'Waiting all phases to complete JKJK')
        await self.wait_all_phases_done()
        uvm_debug(self, 'run_test', 'All phases are done now JKJK')

        #// clean up after ourselves
        #phase_runner_proc.kill()
        l_rs = get_report_server()
        l_rs.report_summarize()
        if self.finish_on_completion:
            # TODO should be linked to cocotb somehow
            self.uvm_report_info('FINISH', '$finish was reached in run_test()', UVM_NONE)


    async def wait_all_phases_done(self):
        await self.m_phase_all_done_event.wait()
        #wait (m_phase_all_done == 1)

    def die(self):
        """
        This method is called by the report server if a report reaches the maximum
        quit count or has a UVM_EXIT action associated with it, e.g., as with
        fatal errors.

        Calls the <uvm_component::pre_abort()> method
        on the entire `uvm_component` hierarchy in a bottom-up fashion.
        It then calls <uvm_report_server::report_summarize> and terminates the simulation.

        Raises:
            Exception:
        """
        l_rs = get_report_server()
        # do the pre_abort callbacks
        self.m_do_pre_abort()
        l_rs.report_summarize()
        #$finish
        if UVMRoot.raise_exception_on_die:
            raise Exception('die(): $finish from UVMRoot')

    def running_test_msg(self, test_name, uvm_test_top):
        if test_name == "":
            self.uvm_report_info("RNTST", "Running test ...", UVM_LOW)
        elif test_name == uvm_test_top.get_type_name():
            self.uvm_report_info("RNTST", "Running test " + test_name + "...", UVM_LOW)
        else:
            self.uvm_report_info("RNTST", ("Running test " + uvm_test_top.get_type_name()
                + " (via factory override for test \"" + test_name + "\")..."), UVM_LOW)

    m_uvm_timeout_overridable = 1
    #  // Function: set_timeout
    #  //
    #  // Specifies the timeout for the simulation. Default is <`UVM_DEFAULT_TIMEOUT>
    #  //
    #  // The timeout is simply the maximum absolute simulation time allowed before a
    #  // ~FATAL~ occurs.  If the timeout is set to 20ns, then the simulation must end
    #  // before 20ns, or a ~FATAL~ timeout will occur.
    #  //
    #  // This is provided so that the user can prevent the simulation from potentially
    #  // consuming too many resources (Disk, Memory, CPU, etc) when the testbench is
    #  // essentially hung.
    #  //
    #  //
    def set_timeout(self, timeout, overridable=1):
        if UVMRoot.m_uvm_timeout_overridable == 0:
            self.uvm_report_info("NOTIMOUTOVR", sv.sformatf(
                "Global timeout setting %0d is not overridable to %0d due to a previous setting.",
                self.phase_timeout, timeout), UVM_NONE)
            return

        UVMRoot.m_uvm_timeout_overridable = overridable
        self.phase_timeout = timeout


    #  //----------------------------------------------------------------------------
    #  // Group: Topology
    #  //----------------------------------------------------------------------------


    def find(self, comp_match):
        """
        Find components from the hierarchy.

        Args:
            comp_match (str): String to match.
        Returns:
            UVMComponent: First component matching the string.
        """
        comp_list = []

        self.find_all(comp_match, comp_list)

        if len(comp_list) > 1:
            uvm_report_warning("MMATCH",
            sv.sformatf("Found %0d components matching '%s'. Returning first match, %0s.",
                      len(comp_list), comp_match,comp_list[0].get_full_name()), UVM_NONE)

        if len(comp_list) == 0:
            uvm_report_warning("CMPNFD",
              "Component matching '", comp_match,
               "' was not found in the list of uvm_components", UVM_NONE)
            return None

        return comp_list[0]

    def find_all(self, comp_match, comps: List[UVMComponent], comp=None):
        """
        Returns the component handle (find) or list of components handles
        (find_all) matching a given string. The string may contain the wildcards,
        * and ?. Strings beginning with '.' are absolute path names. If the optional
        argument comp is provided, then search begins from that component down
        (default=all components).

        Args:
            comp_match (str):
            comps (list):
            comp (UVMComponent):
        """
        if comp is None:
            comp = self
        self.m_find_all_recurse(comp_match, comps, comp)

    def print_topology(self, printer=None):
        """
        Print the verification environment's component topology. The
        `printer` is a `UVMPrinter` object that controls the format
        of the topology printout; a `None` printer prints with the
        default output.

        Args:
            printer (UVMPrinter):
        """
        # s = ""
        if len(self.m_children) == 0:
            self.uvm_report_warning("EMTCOMP", "print_topology - No UVM components to print.",
                    UVM_NONE)
            return

        if printer is None:
            from .uvm_global_vars import uvm_default_printer
            printer = uvm_default_printer

        for c in self.m_children:
            if self.m_children[c].print_enabled:
                printer.print_object("", self.m_children[c])
        self.uvm_report_info("UVMTOP", "UVM testbench topology:\n" + printer.emit(), UVM_NONE)



    def m_find_all_recurse(self, comp_match, comps: List[UVMComponent], comp=None):
        """
          PRIVATE members
         extern function void m_find_all_recurse(string comp_match,
                                                 ref uvm_component comps[$],
                                                 input uvm_component comp=None);
        Args:
            comp_match:
            comps:
            comp:
        """
        child_comp = None

        if comp.has_first_child():
            child_comp = comp.get_first_child()
            while True:
                self.m_find_all_recurse(comp_match, comps, child_comp)
                if not comp.has_next_child():
                    break
                else:
                    child_comp = comp.get_next_child()
        if uvm_is_match(comp_match, comp.get_full_name()) and comp.get_name() != "":
            comps.append(comp)



    #  extern protected function new ()

    def m_add_child(self, child: UVMComponent) -> bool:
        """
        Add child to the top levels array.
        Args:
            child (UVMComponent):
        Returns:
            bool: True if adding child succeeds, False otherwise.
        """
        if super().m_add_child(child):
            if child.get_name() == "uvm_test_top":
                self.top_levels.insert(0, child)
            else:
                self.top_levels.append(child)
            return True
        else:
            return False


    def build_phase(self, phase: UVMPhase) -> None:
        """
         extern function void build_phase(uvm_phase phase)
        Args:
            phase:
        """
        UVMComponent.build_phase(self, phase)
        self.m_set_cl_msg_args()
        self.m_do_verbosity_settings()
        self.m_do_timeout_settings()
        self.m_do_factory_settings()
        self.m_do_config_settings()
        self.m_do_max_quit_settings()
        self.m_do_dump_args()


    def m_do_verbosity_settings(self) -> None:
        """
         extern local function void m_do_verbosity_settings()
        """
        set_verbosity_settings = []
        split_vals = []
        #tmp_verb = 0

        # Retrieve them all into set_verbosity_settings
        self.clp.get_arg_values("+uvm_set_verbosity=", set_verbosity_settings)

        for i in range(len(set_verbosity_settings)):
            uvm_split_string(set_verbosity_settings[i], ",", split_vals)
            if len(split_vals) < 4 or len(split_vals) > 5:
                self.uvm_report_warning("INVLCMDARGS",
                  sv.sformatf(MSG_INVLCMDARGS, set_verbosity_settings[i]), UVM_NONE, "", "")

            # Invalid verbosity
            if self.clp.m_convert_verb(split_vals[2]) == -1:
                self.uvm_report_warning("INVLCMDVERB",
                  sv.sformatf("Invalid verbosity found on the command line for setting '%s'.",
                  set_verbosity_settings[i]), UVM_NONE, "", "")
        #endfunction
        #
        #

    def m_do_timeout_settings(self) -> None:
        """
         extern local function void m_do_timeout_settings()
        """
        timeout_settings = []
        timeout = ""
        timeout_int = 0
        override_spec = ""
        timeout_count = self.clp.get_arg_values("+UVM_TIMEOUT=", timeout_settings)

        if timeout_count == 0:
            return
        else:
            timeout = timeout_settings[0]
            if timeout_count > 1:
                timeout_list = ""
                sep = ""
                for i in range(len(timeout_settings)):
                    if i != 0:
                        sep = "; "
                    timeout_list = timeout_list + sep + timeout_settings[i]

                self.uvm_report_warning("MULTTIMOUT",
                  sv.sformatf(MSG_MULTTIMOUT, timeout_count, timeout, timeout_list), UVM_NONE)

            self.uvm_report_info("TIMOUTSET",
                sv.sformatf("'+UVM_TIMEOUT=%s' provided on the command line is being applied.",
                    timeout), UVM_NONE)

            matches = sv.sscanf(timeout,"%d,%s", timeout_int, override_spec)
            if len(matches) == 2:
                override_spec, timeout_int = matches
            if override_spec == "YES":
                self.set_timeout(timeout_int, 1)
            elif override_spec == "NO":
                self.set_timeout(timeout_int, 0)
            else:
                self.set_timeout(timeout_int, 1)


    def m_do_factory_settings(self) -> None:
        """
         extern local function void m_do_factory_settings()
        """
        args = []

        self.clp.get_arg_matches("/^\\+(UVM_SET_INST_OVERRIDE|uvm_set_inst_override)=/",args)
        for i in range(len(args)):
            value = args[i][23:len(args[i])]
            self.m_process_inst_override(value)

        self.clp.get_arg_matches("/^\\+(UVM_SET_TYPE_OVERRIDE|uvm_set_type_override)=/",args)
        for i in range(len(args)):
            value = args[i][23:len(args[i])]
            self.m_process_type_override(value)


    def m_process_inst_override(self, ovr) -> None:
        """
        extern local function void m_process_inst_override(string ovr)
        Args:
            ovr:
        """
        from .uvm_coreservice import UVMCoreService
        split_val = []
        cs = UVMCoreService.get()
        factory = cs.get_factory()

        uvm_split_string(ovr, ",", split_val)

        if len(split_val) != 3:
            self.uvm_report_error("UVM_CMDLINE_PROC", "Invalid setting for +uvm_set_inst_override="
                + ovr
                + ", setting must specify <requested_type>,<override_type>,<instance_path>", UVM_NONE)
            return

        uvm_info("INSTOVR",
            "Applying instance override from the command line: +uvm_set_inst_override=" + str(ovr),
            UVM_NONE)
        factory.set_inst_override_by_name(split_val[0], split_val[1], split_val[2])
        #endfunction
        #

    def m_process_type_override(self, ovr):
        """
         extern local function void m_process_type_override(string ovr)
        Args:
            ovr:
        """
        pass  # TODO


    def m_do_config_settings(self):
        """
        Processes config value options set from cmdline:
          +uvm_set_config_int=
          +uvm_set_config_string=
        """
        args = []

        self.clp.get_arg_matches("/^\\+(UVM_SET_CONFIG_INT|uvm_set_config_int)=/", args)
        for i in range(len(args)):
            self.m_process_config(args[i][20:len(args[i])], 1)

        self.clp.get_arg_matches("/^\\+(UVM_SET_CONFIG_STRING|uvm_set_config_string)=/",args)
        for i in range(len(args)):
            self.m_process_config(args[i][23:len(args[i])], 0)

        # TODO might not be needed, ever
        #self.clp.get_arg_matches("/^\\+(UVM_SET_DEFAULT_SEQUENCE|uvm_set_default_sequence)=/", args)
        #for i in range(len(args)):
        #    self.m_process_default_sequence(args[i][26:len(args[i])])


    #  extern local function void m_do_max_quit_settings()
    def m_do_max_quit_settings(self):
        max_quit_settings = []
        max_quit_count = 0
        max_quit = ""
        split_max_quit = []
        max_quit_int = 0
        srvr = UVMReportServer.get_server()
        max_quit_count = self.clp.get_arg_values("+UVM_MAX_QUIT_COUNT=", max_quit_settings)
        if max_quit_count == 0:
            return
        else:
            max_quit = max_quit_settings[0]
            if max_quit_count > 1:
                max_quit_list = ""
                sep = ""
                for i in range(len(max_quit_settings)):
                    if i != 0:
                        sep = "; "
                    max_quit_list = max_quit_list + sep + max_quit_settings[i]

                self.uvm_report_warning("MULTMAXQUIT",
                  sv.sformatf(WARN_MAX_QUIT_COUNT,
                  max_quit_count, max_quit, max_quit_list), UVM_NONE)

            self.uvm_report_info("MAXQUITSET",
                sv.sformatf("'+UVM_MAX_QUIT_COUNT=%s' provided on the command line is being applied.",
                    max_quit), UVM_NONE)
            uvm_split_string(max_quit, ",", split_max_quit)
            max_quit_int = int(split_max_quit[0])
            if len(split_max_quit) > 1:
                if split_max_quit[1] == "YES":
                    srvr.set_max_quit_count(max_quit_int, 1)
                elif split_max_quit[1] == "NO":
                    srvr.set_max_quit_count(max_quit_int, 0)
            else:
                srvr.set_max_quit_count(max_quit_int, 1)

    def m_do_dump_args(self):
        """
         extern local function void m_do_dump_args()
        """
        dump_args = []
        all_args = []
        out_string = ""
        if self.clp.get_arg_matches("+UVM_DUMP_CMDLINE_ARGS", dump_args):
            all_args = self.clp.get_args()
            for i in range(len(all_args)):
                if (all_args[i] == "__-f__"):
                    continue
                out_string = out_string + all_args[i] + " "
            uvm_info("DUMPARGS", out_string, UVM_NONE)


    def m_process_config(self, cfg, is_int):
        """
        extern local function void m_process_config(string cfg, bit is_int)
        Args:
            cfg:
            is_int:
        """
        from .uvm_coreservice import UVMCoreService
        v = 0
        split_val = []
        cs = UVMCoreService.get()
        m_uvm_top = cs.get_root()

        uvm_split_string(cfg, ",", split_val)
        if len(split_val) == 1:
            uvm_report_error("UVM_CMDLINE_PROC", ("Invalid +uvm_set_config command\""
                + cfg + "\" missing field and value: component is \""
                + split_val[0] + "\""), UVM_NONE)
            return

        if len(split_val) == 2:
            uvm_report_error("UVM_CMDLINE_PROC", "Invalid +uvm_set_config command\""
                + cfg + "\" missing value: component is \"" + split_val[0]
                + "\"  field is \"" + split_val[1] + "\"", UVM_NONE)
            return

        if len(split_val) > 3:
            uvm_report_error("UVM_CMDLINE_PROC", sv.sformatf(
                "Invalid +uvm_set_config command\"%s\" : expected only 3 fields (component, field and value).",
                cfg), UVM_NONE)
            return

        if is_int:
            if len(split_val[2]) > 2:
                #base = split_val[2].substr(0,1)
                base = split_val[2][0:2]
                extval = split_val[2][2:len(split_val[2])]
                if base in ["'b", "0b"]:
                    v = int(extval, 2)
                elif base == "'o":
                    v = int(extval, 8)
                elif base == "'d":
                    v = int(extval, 10)
                elif base in ["'h", "'x", "0x"]:
                    v = int(extval, 16)
                else:
                    v = int(split_val[2])
            else:
                v = int(split_val[2])

            self.uvm_report_info("UVM_CMDLINE_PROC", ("Applying config setting "
                    + "from the command line: +uvm_set_config_int=" + cfg), UVM_NONE)
            UVMConfigDb.set(m_uvm_top, split_val[0], split_val[1], v)
        else:
            self.uvm_report_info("UVM_CMDLINE_PROC",
                "Applying config setting from the command line: +uvm_set_config_string=" + cfg, UVM_NONE)
            UVMConfigDb.set(m_uvm_top, split_val[0], split_val[1], split_val[2])


    #  extern local function void m_process_default_sequence(string cfg)

    def m_check_verbosity(self):
        """
         extern function void m_check_verbosity()
        """
        verb_string = ""
        verb_settings = []
        verb_count = 0
        plusarg = 0
        verbosity = UVM_MEDIUM

        # Retrieve the verbosities provided on the command line.
        verb_count = self.clp.get_arg_values("+UVM_VERBOSITY=", verb_settings)
        if verb_count:
            verb_settings.append(verb_string)

        # If none provided, provide message about the default being used.
        if verb_count == 0:
            self.uvm_report_info("DEFVERB",
                "No verbosity specified on the command line.  Using the default: UVM_MEDIUM", UVM_NONE)
        #
        # If at least one, use the first.
        if verb_count > 0:
            verb_string = verb_settings[0]
            plusarg = 1

        # If more than one, provide the warning stating how many, which one will
        # be used and the complete list.
        if (verb_count > 1):
            verb_list = ""
            sep = ""
            for i in range(len(verb_settings)):
                if (i != 0):
                    sep = ", "
                verb_list = verb_list + sep + verb_settings[i]
            self.uvm_report_warning("MULTVERB",
              sv.sformatf(WARN_MSG1, verb_count, verb_string, verb_list), UVM_NONE)

        if (plusarg == 1):
            # case(verb_string)
            if verb_string == "UVM_NONE":
                verbosity = UVM_NONE
            elif verb_string == "NONE":
                verbosity = UVM_NONE
            elif verb_string == "UVM_LOW":
                verbosity = UVM_LOW
            elif verb_string == "LOW":
                verbosity = UVM_LOW
            elif verb_string == "UVM_MEDIUM":
                verbosity = UVM_MEDIUM
            elif verb_string == "MEDIUM":
                verbosity = UVM_MEDIUM
            elif verb_string == "UVM_HIGH":
                verbosity = UVM_HIGH
            elif verb_string == "HIGH":
                verbosity = UVM_HIGH
            elif verb_string == "UVM_FULL":
                verbosity = UVM_FULL
            elif verb_string == "FULL":
                verbosity = UVM_FULL
            elif verb_string == "UVM_DEBUG":
                verbosity = UVM_DEBUG
            elif verb_string == "DEBUG":
                verbosity = UVM_DEBUG
            else:
                verbosity = int(verb_string)
                if verbosity > 0:
                    self.uvm_report_info("NSTVERB",
                            sv.sformatf(NON_STD_VERB, verbosity), UVM_NONE)
                if verbosity == 0:
                    verbosity = UVM_MEDIUM
                    self.uvm_report_warning("ILLVERB", "Illegal verbosity value, using default of UVM_MEDIUM.", UVM_NONE)

        self.set_report_verbosity_level_hier(verbosity)

        #endfunction

    def report_header(self, _file=0) -> None:
        """
         extern virtual function void report_header(UVM_FILE file = 0)
        Args:
            _file:
        """
        q = []
        args = []

        #srvr = UVMReportServer.get_server()
        clp = UVMCmdlineProcessor.get_inst()

        if clp.get_arg_matches("+UVM_NO_RELNOTES", args):
            return

        q.append("\n" + LINE + "\n")
        q.append(uvm_revision_string() + "\n")
        q.append(uvm_mgc_copyright + "\n")
        q.append(uvm_cdn_copyright + "\n")
        q.append(uvm_snps_copyright + "\n")
        q.append(uvm_cy_copyright + "\n")
        q.append(uvm_nv_copyright + "\n")
        q.append(uvm_tpoikela_copyright + "\n")
        q.append(LINE + "\n")

        if UVM_OBJECT_DO_NOT_NEED_CONSTRUCTOR == 0:
            if UVMRoot.m_relnotes_done is False:
                q.append("\n  ***********       IMPORTANT RELEASE NOTES         ************\n")
            q.append("\n  You are using a Python version of the UVM library which \n")
            q.append("  requires cocotb and an HDL simulator.\n")
            UVMRoot.m_relnotes_done = True

        if UVMRoot.m_relnotes_done is True:
            q.append("\n      (Specify +UVM_NO_RELNOTES to turn off this notice)\n")

        self.uvm_report_info("UVM/RELNOTES", UVM_STRING_QUEUE_STREAMING_PACK(q), UVM_LOW)


    #  // For error checking
    #  extern virtual task run_phase (uvm_phase phase)

    def phase_started(self, phase) -> None:
        """
        At end of elab phase we need to do tlm binding resolution.

        Args:
            phase (UVMPhase): Current phase for this callback.
        """
        eof_elab_phase = UVMEndOfElaborationPhase.get()
        uvm_debug(self, 'phase_started', phase.get_name())
        #if phase == end_of_elaboration_ph:
        if phase.get_name() == eof_elab_phase.get_name():
            uvm_debug(self, 'phase_started', "uvm_root resolving bindings now..")
            self.do_resolve_bindings()
            if self.enable_print_topology:
                self.print_topology()
            srvr = UVMReportServer.get_server()
            if srvr.get_severity_count(UVM_ERROR) > 0:
                self.uvm_report_fatal("BUILDERR", "stopping due to build errors", UVM_NONE)


    # function void end_of_elaboration_phase(uvm_phase phase);
    #	 uvm_component_proxy p = new("proxy")
    #	 uvm_top_down_visitor_adapter#(uvm_component) adapter = new("adapter")
    #	 uvm_coreservice_t cs = uvm_coreservice_t::get()
    #	 uvm_visitor#(uvm_component) v = cs.get_component_visitor()
    #	 adapter.accept(this, v, p)
    # endfunction


#----------------------------------------------------------------------------
# Group: Global Variables
#----------------------------------------------------------------------------

def get_report_server() -> 'UVMReportServer':
    from .uvm_report_server import UVMReportServer
    return UVMReportServer.get_server()

#------------------------------------------------------------------------------
# Variable: uvm_top
#
# This is the top-level that governs phase execution and provides component
# search interface. See <uvm_root> for more information.
#------------------------------------------------------------------------------
uvm_top: UVMRoot = UVMRoot.get()

#//-----------------------------------------------------------------------------
#// IMPLEMENTATION
#//-----------------------------------------------------------------------------
#
#
#
#
#// m_process_type_override
#// -----------------------
#
#function void uvm_root::m_process_type_override(string ovr)
#  string split_val[$]
#  int replace=1
#  uvm_coreservice_t cs = uvm_coreservice_t::get();
#  uvm_factory factory=cs.get_factory()
#
#  uvm_split_string(ovr, ",", split_val)
#
#  if(split_val.size() > 3 || split_val.size() < 2):
#    uvm_report_error("UVM_CMDLINE_PROC", {"Invalid setting for +uvm_set_type_override=", ovr,
#      ", setting must specify <requested_type>,<override_type>[,<replace>]"}, UVM_NONE)
#    return
#  end
#
#  // Replace arg is optional. If set, must be 0 or 1
#  if(split_val.size() == 3):
#    if(split_val[2]=="0") replace =  0
#    else if (split_val[2] == "1") replace = 1
#    else begin
#      uvm_report_error("UVM_CMDLINE_PROC", {"Invalid replace arg for +uvm_set_type_override=", ovr ," value must be 0 or 1"}, UVM_NONE)
#      return
#    end
#  end
#
#  self.uvm_report_info("UVM_CMDLINE_PROC", {"Applying type override from the command line: +uvm_set_type_override=", ovr}, UVM_NONE)
#  factory.set_type_override_by_name(split_val[0], split_val[1], replace)
#endfunction
#
#
#
#// m_process_default_sequence
#// ----------------
#
#function void uvm_root::m_process_default_sequence(string cfg)
#  string split_val[$]
#  uvm_coreservice_t cs = uvm_coreservice_t::get()
#  uvm_root m_uvm_top = cs.get_root();
#  uvm_factory f = cs.get_factory()
#  uvm_object_wrapper w
#
#  uvm_split_string(cfg, ",", split_val)
#  if(split_val.size() == 1):
#    uvm_report_error("UVM_CMDLINE_PROC", {"Invalid +uvm_set_default_sequence command\"", cfg,
#      "\" missing phase and type: sequencer is \"", split_val[0], "\""}, UVM_NONE)
#    return
#  end
#
#  if(split_val.size() == 2):
#    uvm_report_error("UVM_CMDLINE_PROC", {"Invalid +uvm_set_default_sequence command\"", cfg,
#      "\" missing type: sequencer is \"", split_val[0], "\"  phase is \"", split_val[1], "\""}, UVM_NONE)
#    return
#  end
#
#  if(split_val.size() > 3):
#    uvm_report_error("UVM_CMDLINE_PROC",
#      $sformatf("Invalid +uvm_set_default_sequence command\"%s\" : expected only 3 fields (sequencer, phase and type).", cfg), UVM_NONE)
#    return
#  end
#
#  w = f.find_wrapper_by_name(split_val[2])
#  if (w == None):
#      uvm_report_error("UVM_CMDLINE_PROC",
#                       $sformatf("Invalid type '%s' provided to +uvm_set_default_sequence", split_val[2]),
#                       UVM_NONE)
#      return
#  end
#  else begin
#      self.uvm_report_info("UVM_CMDLINE_PROC", {"Setting default sequence from the command line: +uvm_set_default_sequence=", cfg}, UVM_NONE)
#      uvm_config_db#(uvm_object_wrapper)::set(this, {split_val[0], ".", split_val[1]}, "default_sequence", w)
#  end
#
#endfunction : m_process_default_sequence
#
#
#
#
#// It is required that the run phase start at simulation time 0
#// TBD this looks wrong - taking advantage of uvm_root not doing anything else?
#// TBD move to phase_started callback?
#task uvm_root::run_phase (uvm_phase phase)
#  // check that the commandline are took effect
#  foreach(m_uvm_applied_cl_action[idx])
#	  if(m_uvm_applied_cl_action[idx].used==0):
#		    `uvm_warning("INVLCMDARGS",$sformatf("\"+uvm_set_action=%s\" never took effect due to a mismatching component pattern",m_uvm_applied_cl_action[idx].arg))
#	  end
#  foreach(m_uvm_applied_cl_sev[idx])
#	  if(m_uvm_applied_cl_sev[idx].used==0):
#		    `uvm_warning("INVLCMDARGS",$sformatf("\"+uvm_set_severity=%s\" never took effect due to a mismatching component pattern",m_uvm_applied_cl_sev[idx].arg))
#	  end
#
#  if($time > 0)
#    `uvm_fatal("RUNPHSTIME", {"The run phase must start at time 0, current time is ",
#       $sformatf("%0t", $realtime), ". No non-zero delays are allowed before ",
#       "run_test(), and pre-run user defined phases may not consume ",
#       "simulation time before the start of the run phase."})
#endtask



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010-2011 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import Timer, RisingEdge

from uvm.macros import *
from uvm.base import sv, UVMConfigDb, UVMComponent
from uvm.reg import UVMRegPredictor

from regmodel import dut_regmodel
from apb.apb_agent import apb_agent
from apb.apb_rw import reg2apb_adapter

EXPLICIT_MON = 1

#//
#// This example shows how to integrate a register model
#// directly onto a bus sequencer.
#//
#// By default, the mirror in the register model is updated implicitly.
#// For explicit monitoring, define the `EXPLICIT_MON macro
#//


class tb_env(UVMComponent):


    def __init__(self, name, parent=None):
        super().__init__(name,parent)
        self.regmodel = None  # dut_regmodel
        self.Aapb = None  # apb_agent
        self.seq = None  # uvm_reg_sequence
        self.vif = None
        if EXPLICIT_MON:
            self.apb2reg_predictor = None  # uvm_reg_predictor#(apb_rw)


    def build_phase(self, phase):
        if self.regmodel is None:
            self.regmodel = dut_regmodel.type_id.create("regmodel",None,self.get_full_name())
            self.regmodel.build()
            self.regmodel.lock_model()

            self.apb = apb_agent.type_id.create("apb", self)
        if EXPLICIT_MON:
            self.apb2reg_predictor = UVMRegPredictor("apb2reg_predictor", self)

        hdl_root = "dut"
        sv.value_plusargs("ROOT_HDL_PATH=%s",hdl_root)  # cast to 'void' removed
        self.regmodel.set_hdl_path_root(hdl_root)
        vif = []
        if UVMConfigDb.get(self, "apb", "vif", vif):
            self.vif = vif[0]
        else:
            uvm_fatal("NO_VIF", "Could not find vif from config DB")


    def connect_phase(self, phase):
        if self.apb is not None:
            self.reg2apb = reg2apb_adapter('adapter')
            self.regmodel.default_map.set_sequencer(self.apb.sqr, self.reg2apb)
            if EXPLICIT_MON:
                self.apb2reg_predictor.map = self.regmodel.default_map
                self.apb2reg_predictor.adapter = self.reg2apb
                self.regmodel.default_map.set_auto_predict(0)
                self.apb.mon.ap.connect(self.apb2reg_predictor.bus_in)
            else:
                self.regmodel.default_map.set_auto_predict(1)
            self.regmodel.print_obj()


    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        if self.seq is None:
            uvm_fatal("NO_SEQUENCE","Env's sequence is not defined. Nothing to do. Exiting.")
            return

        # begin : do_reset
        uvm_info("RESET","Performing reset of 5 cycles", UVM_LOW)
        self.vif.rst <= 1
        for _ in range(5):
            await RisingEdge(self.vif.clk)
        self.vif.rst <= 0

        await Timer(100, "NS")

        uvm_info("START_SEQ", "Starting sequence '" + self.seq.get_name() + "'", UVM_LOW)
        self.seq.model = self.regmodel
        await self.seq.start(None)
        phase.drop_objection(self)


uvm_component_utils(tb_env)



#//----------------------------------------------------------------------
#//   Copyright 2010 Mentor Graphics Corporation
#//   Copyright 2010-2011 Synopsys, Inc
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#

from uvm import UVMComponent, uvm_component_utils
from uvm.macros import *

from initiator import initiator
from target import target

class tb_env(UVMComponent):

    def __init__(self, name="tb_env", parent=None):
        super().__init__(name, parent)
        self.master = None
        self.slave = None

    def build_phase(self, phase):
        self.master = initiator.type_id.create("master", self)
        self.slave  = target.type_id.create("slave", self)

    def connect_phase(self, phase):
        self.master.sock.connect(self.slave.sock)

    def check_phase(self, phase):
        if self.slave.num_transport == 0:
            uvm_fatal("ERR_NUM_TRANSPORT", "Got zero items!")

uvm_component_utils(tb_env)




import unittest
from uvm.reg.uvm_reg import UVMReg
from uvm.reg.uvm_reg_field import UVMRegField
from uvm.uvm_unit import create_reg_map


class TestUVMReg(unittest.TestCase):

    def setUp(self):
        pass

    def create_reg(self, name='reg', nbits=32):
        reg = UVMReg(name, nbits, False)
        for i in range(0, nbits):
            f1 = UVMRegField('my_field_' + str(i))
            f1.configure(reg, 1, i, 'RW', volatile=False, reset=i % 2, has_reset=True,
                    is_rand=False, individually_accessible=True)
        return reg

    def test_add_field(self):
        UVMRegField.define_access('RW')
        reg = UVMReg('my_reg', 32, False)
        f1 = UVMRegField('my_field_1')
        f1.configure(reg, 16, 0, 'RW', volatile=False, reset=0, has_reset=True,
                is_rand=False, individually_accessible=True)
        arr = []
        reg.get_fields(arr)
        print(str(reg.m_fields))
        self.assertEqual(len(arr), 1)

    def test_base_ops(self):
        UVMRegField.define_access('RW')
        reg32 = self.create_reg('reg_32', 32)
        arr = []
        reg32.get_fields(arr)
        self.assertEqual(len(arr), 32)
        self.assertEqual(reg32.predict(0), True)
        # Why does not fail self.assertEqual(reg32.predict(1234), False)
        self.assertEqual(reg32.get(), 0)
        reg32.set(0x12345678)
        val = reg32.get()
        self.assertEqual(val, 0x12345678)

    def test_reg_randomize(self):
        class RandReg(UVMReg):
            def __init__(self, name):
                super().__init__(name, 16)
                self.F1 = UVMRegField('my_field_1')
                self.F1.configure(self, 16, 0, 'RW', volatile=False, reset=0,
                        has_reset=True, is_rand=False, individually_accessible=True)
                self.rand('F1')

        new_reg = RandReg('rng_reg')
        ok = new_reg.randomize()
        self.assertTrue(ok, 'randomize OK')
        self.assertTrue(new_reg.F1.value > 0)

    def test_get_address(self):
        reg = self.create_reg('my_reg', 32)
        reg_map, rb = create_reg_map('test_map')
        reg.configure(rb)
        reg_map.add_reg(reg, 0x8)
        # TODO fix this addr = reg.get_address()
        # self.assertEqual(addr, 0x8)


if __name__ == '__main__':
    unittest.main()




import unittest
from uvm.tlm1.uvm_tlm_fifos import (UVMTLMAnalysisFIFO, UVMTLMFIFO)


class TestUVMTLMFIFO(unittest.TestCase):

    def test_flush(self):
        fifo = UVMTLMFIFO('tlm_fifo', None)
        fifo.flush()
        self.assertEqual(fifo.used(), 0)
        self.assertTrue(fifo.is_empty())


class TestUVMTLMAnalysisFIFO(unittest.TestCase):

    def test_write_and_get(self):
        fifo = UVMTLMAnalysisFIFO('fifo', None)
        fifo.write(12345)
        arr = []
        self.assertEqual(fifo.try_get(arr), True)
        self.assertEqual(arr[0], 12345)
        self.assertEqual(fifo.can_get(), False)
        self.assertEqual(fifo.try_put('xxx'), True)
        self.assertEqual(fifo.can_get(), True)
        arr = []
        self.assertEqual(fifo.try_get(arr), True)
        self.assertEqual(arr[0], 'xxx')

    def test_try_put(self):
        fifo = UVMTLMAnalysisFIFO('fifo2', None)
        for i in range(10):
            ok = fifo.try_put(i * 2 + 1)
            self.assertTrue(ok)
        for i in range(10):
            arr = []
            ok = fifo.try_get(arr)
            self.assertEqual(arr[0], i * 2 + 1)



#//
#// -------------------------------------------------------------
#//    Copyright 2010-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2019 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm.macros import uvm_object_utils
from uvm.reg import *
from uvm.base import *

#//
#// This example demonstrates how to include a user-defined register
#// in a register model.
#//

#//
#// The user_acp_reg has a user-defined behavior
#//
#// It increments by 1 after every write
#//

class user_acp_incr_on_write_cbs(UVMRegCbs):

    def __init__(self, name="write_cbs"):
        UVMRegCbs.__init__(self, name)
        self.num_called = 0


    def post_predict(self, fld, previous, value, kind, path, _map):
        if (kind != UVM_PREDICT_WRITE):
            return
        if (path != UVM_FRONTDOOR):
            return
        value.push(previous + 1)
        self.num_called += 1


class user_acp_reg(UVMReg):
    #
    #   local UVMRegField value;
    #
    #
    def __init__(self, name="user_acp_reg"):
        UVMReg.__init__(self, name,16,UVM_NO_COVERAGE)

    def build(self):
        self.value = UVMRegField.type_id.create("value", None, self.get_full_name())
        self.value.configure(self, 16, 0, "RW", 0, 0x0000, 1, 0, 0);
        self.value.set_compare(UVM_NO_CHECK);

        UVMResourceDb.set("REG::" + self.get_full_name(),
                "NO_REG_BIT_BASH_TEST", 1)
        UVMResourceDb.set("REG::" + self.get_full_name(),
                "NO_REG_ACCESS_TEST", 1);
        self.cb = user_acp_incr_on_write_cbs()
        UVMRegFieldCb.add(self.value, self.cb)
        #   endfunction: build


    
    async def pre_write(self, rw):
        m_data = 0
        rg = None

        #assert($cast(rg,rw.element));
        rg = rw.element

        # Predict the value that will be in the register
        m_data = rg.get() + 1

        # If a backdoor write is used, replace the value written
        # with the incremented value to emulate the front-door
        if (rw.path == UVM_BACKDOOR):
            rw.value[0] = m_data
        await Timer(5, "NS")
        #   endtask: pre_write
    #
    #endclass : user_acp_reg

uvm_object_utils(user_acp_reg)


class block_B(UVMRegBlock):
    #   user_acp_reg user_acp;

    def __init__(self, name="B"):
        UVMRegBlock.__init__(self, name, UVM_NO_COVERAGE)

    def build(self):
        self.default_map = self.create_map("", 0, 1, UVM_BIG_ENDIAN)
        self.user_acp = user_acp_reg.type_id.create("user_acp", None, self.get_full_name())
        self.user_acp.configure(self, None, "acp")
        self.user_acp.build()
    
        self.default_map.add_reg(self.user_acp, 0x0000, "RW")
        self.lock_model()

    #endclass : block_B
uvm_object_utils(block_B)



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

from uvm.base.uvm_component import UVMComponent
from uvm.tlm1 import UVMAnalysisImp


class UVMSubscriber(UVMComponent):  # (type T=int) extends uvm_component
    """
    This class provides an analysis export for receiving transactions from a
    connected analysis export. Making such a connection "subscribes" this
    component to any transactions emitted by the connected analysis port.

    Subtypes of this class must define the write method to process the incoming
    transactions. This class is particularly useful when designing a coverage
    collector that attaches to a monitor.

    :ivar UVMAnalysisImp analysis_export: This export provides access to the
        write method, which derived subscribers must implement.
    """


    #  // Function: new
    #  //
    #  // Creates and initializes an instance of this class using the normal
    #  // constructor arguments for <uvm_component>: ~name~ is the name of the
    #  // instance, and ~parent~ is the handle to the hierarchical parent, if any.
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.analysis_export = UVMAnalysisImp("analysis_imp", self)

    #  // Function: write
    #  //
    #  // A method that must be defined in each subclass. Access
    #  // to this method by outside components should be done via the
    #  // analysis_export.
    def write(self, t):
        raise Exception("Pure virtual function. Must be implemented")



#
from .uvm_analysis_port import *
from .uvm_exports import *
from .uvm_imps import *
from .uvm_ports import *
from .uvm_sqr_connections import *
from .uvm_sqr_ifs import *
from .uvm_tlm_fifo_base import *
from .uvm_tlm_fifos import *
from .uvm_tlm_ifs import *
from .uvm_tlm_imps import *
from .uvm_tlm import *
from .uvm_tlm_req_rsp import *



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer, RisingEdge

from uvm.comps import UVMMonitor
from ubus_transfer import *
from uvm.tlm1 import UVMAnalysisPort, UVMBlockingPeekImp
from uvm.base import *

#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_master_monitor
#//
#//------------------------------------------------------------------------------

#class ubus_master_monitor extends uvm_monitor
class ubus_master_monitor(UVMMonitor):
    #
    #  // This property is the virtual interfaced needed for this component to drive
    #  // and view HDL signals. 
    #  protected virtual ubus_if vif
    #
    #  // The following two bits are used to control whether checks and coverage are
    #  // done both in the monitor class and the interface.
    #  bit checks_enable = 1
    #  bit coverage_enable = 1
    #
    #  uvm_analysis_port #(ubus_transfer) item_collected_port
    #
    #  // The following property holds the transaction information currently
    #  // begin captured (by the collect_address_phase and data_phase methods). 
    #  protected ubus_transfer trans_collected
    #
    #  // Fields to hold trans addr, data and wait_state.
    #  protected bit [15:0] addr
    #  protected bit [7:0] data
    #  protected int unsigned wait_state
    #
    #  // Transfer collected covergroup
    #  covergroup cov_trans
    #    option.per_instance = 1
    #    trans_start_addr : coverpoint trans_collected.addr {
    #      option.auto_bin_max = 16; }
    #    trans_dir : coverpoint trans_collected.read_write
    #    trans_size : coverpoint trans_collected.size {
    #      bins sizes[] = {1, 2, 4, 8}
    #      illegal_bins invalid_sizes = default; }
    #    trans_addrXdir : cross trans_start_addr, trans_dir
    #    trans_dirXsize : cross trans_dir, trans_size
    #  endgroup : cov_trans
    #
    #  // Transfer collected beat covergroup
    #  covergroup cov_trans_beat
    #    option.per_instance = 1
    #    beat_addr : coverpoint addr {
    #      option.auto_bin_max = 16; }
    #    beat_dir : coverpoint trans_collected.read_write
    #    beat_data : coverpoint data {
    #      option.auto_bin_max = 8; }
    #    beat_wait : coverpoint wait_state {
    #      bins waits[] = { [0:9] }
    #      bins others = { [10:$] }; }
    #    beat_addrXdir : cross beat_addr, beat_dir
    #    beat_addrXdata : cross beat_addr, beat_data
    #  endgroup : cov_trans_beat
    #
    #  // Provide implementations of virtual methods such as get_type_name and create
    #  `uvm_component_utils_begin(ubus_master_monitor)
    #    `uvm_field_int(master_id, UVM_DEFAULT)
    #    `uvm_field_int(checks_enable, UVM_DEFAULT)
    #    `uvm_field_int(coverage_enable, UVM_DEFAULT)
    #  `uvm_component_utils_end
    #

    def __init__(self, name, parent=None):
        UVMMonitor.__init__(self, name, parent)
        #    cov_trans = new()
        #    cov_trans.set_inst_name({get_full_name(), ".cov_trans"})
        #    cov_trans_beat = new()
        #    cov_trans_beat.set_inst_name({get_full_name(), ".cov_trans_beat"})
        self.trans_collected = ubus_transfer()
        self.item_collected_port = UVMAnalysisPort("item_collected_port", self)
        self.addr_ph_imp = UVMBlockingPeekImp("addr_ph_imp", self)
        # The following two bits are used to control whether checks and coverage are
        # done both in the monitor class and the interface.
        self.checks_enable = True
        self.coverage_enable = True
        self.vif = None
        self.master_id = 0
        #  endfunction : new


    def build_phase(self, phase):
        arr = []
        if UVMConfigDb.get(self, "", "vif", arr):
            self.vif = arr[0]
        if self.vif is None:
            self.uvm_report_fatal("NOVIF", "virtual interface must be set for: " +
                self.get_full_name() + ".vif")
        arr = []
        if UVMConfigDb.get(self, "", "master_id", arr):
            self.master_id = arr[0]


    
    async def run_phase(self, phase):
        self.uvm_report_info(self.get_full_name() + " MASTER ID", sv.sformatf(" = %0d",
            self.master_id), UVM_MEDIUM)
        #    fork
        forked_proc = cocotb.fork(self.collect_transactions())
        #    join
        await forked_proc
        #  endtask : run_phase

    #  // collect_transactions
    
    async def collect_transactions(self):
        await Timer(0, "NS")
        while True:
            await RisingEdge(self.vif.sig_clock)
            if (self.m_parent is not None):
                self.trans_collected.master = self.m_parent.get_name()
            await self.collect_arbitration_phase()
            await self.collect_address_phase()
            await self.collect_data_phase()
            uvm_info(self.get_full_name(), sv.sformatf("Transfer collected :\n%s",
                self.trans_collected.sprint()), UVM_MEDIUM)
            if (self.checks_enable):
                self.perform_transfer_checks()
            if (self.coverage_enable):
                self.perform_transfer_coverage()
            self.item_collected_port.write(self.trans_collected)

        #  endtask : collect_transactions

    #  // collect_arbitration_phase
    
    async def collect_arbitration_phase(self):
        #    @(posedge vif.sig_request[master_id])
        #    @(posedge vif.sig_clock iff vif.sig_grant[master_id] === 1)
        while True:
            await RisingEdge(self.vif.sig_request)
            sig_req = self.vif.sig_req[self.master_id]
            if sig_req == 1:
                break

        while True:
            await RisingEdge(self.vif.sig_clock)
            grant = int(self.vif.sig_gnt[self.master_id])
            if grant == 1:
                break
        self.begin_tr(self.trans_collected)
        #  endtask : collect_arbitration_phase

    #  // collect_address_phase
    #  virtual protected task collect_address_phase()
    
    async def collect_address_phase(self):
        await RisingEdge(self.vif.sig_clock)
        self.trans_collected.addr = int(self.vif.sig_addr.value)
        sig_size = int(self.vif.sig_size)

        #case (vif.sig_size)
        if sig_size == 0:
            self.trans_collected.size = 1
        elif sig_size == 1:
            self.trans_collected.size = 2
        elif sig_size == 2:
            self.trans_collected.size = 4
        elif sig_size == 3:
            self.trans_collected.size = 8
        self.trans_collected.data = [0] * self.trans_collected.size

        sig_read = int(self.vif.sig_read)
        sig_write = int(self.vif.sig_write)
        read_write = sig_read << 1 | sig_write
        if read_write == 0:
            self.trans_collected.read_write = NOP
        elif read_write == 2:
            self.trans_collected.read_write = READ
        elif read_write == 1:
            self.trans_collected.read_write = WRITE
        #  endtask : collect_address_phase


    #  // collect_data_phase
    
    async def collect_data_phase(self):
        if (self.trans_collected.read_write != NOP):
            for i in range(self.trans_collected.size):
                while True:
                    await RisingEdge(self.vif.sig_clock)
                    if self.vif.sig_wait.value == 0:
                        break
                self.trans_collected.data[i] = self.vif.sig_data.value
        self.end_tr(self.trans_collected)
    #  endtask : collect_data_phase


    def perform_transfer_checks(self):
        self.check_transfer_size()
        self.check_transfer_data_size()


    #  // check_transfer_size
    def check_transfer_size(self):
        trans_collected = self.trans_collected
        if (trans_collected.size == 1 or
                trans_collected.size == 2 or trans_collected.size == 4 or
                trans_collected.size == 8):
            pass
        else:
            uvm_error(self.get_type_name(), "Invalid transfer size!")


    #  // check_transfer_data_size
    def check_transfer_data_size(self):
        if (self.trans_collected.size != len(self.trans_collected.data)):
            uvm_error(self.get_type_name(), "Transfer size field / data size mismatch.")


    #  // perform_transfer_coverage
    #  virtual protected function void perform_transfer_coverage()
    #    cov_trans.sample()
    #    for (int unsigned i = 0; i < trans_collected.size; i++) begin
    #      addr = trans_collected.addr + i
    #      data = trans_collected.data[i]
    #//Wait state is not currently monitored
    #//      wait_state = trans_collected.wait_state[i]
    #      cov_trans_beat.sample()
    #    end
    #  endfunction : perform_transfer_coverage
    def perform_transfer_coverage(self):
        pass  # TODO

    #
    #  virtual function void report_phase(uvm_phase phase)
    #    `uvm_info(get_full_name(),$sformatf("Covergroup 'cov_trans' coverage: %2f",
    #					cov_trans.get_inst_coverage()),UVM_LOW)
    #  endfunction
    #
    #endclass : ubus_master_monitor
uvm_component_utils(ubus_master_monitor)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2019 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import Timer

from uvm.comps import UVMEnv
from uvm.macros import uvm_component_utils

from regmodel import block_B
from common import reg_agent, reg2rw_adapter
from uvm.reg.uvm_reg_predictor import UVMRegPredictor


class dut_rw():

    @classmethod
    
    async def rw(rw, dut):
        if (rw.addr == 0x0000):
            #'h0000: // acp (MSB)
            if (rw.read):
                rw.data = dut.acp[15:8]
        elif (rw.addr == 0x0001):  # 'h0001: // acp (LSB)
            if (rw.read):
                rw.data = dut.acp[7:0]
            else:
                dut.acp += 1
        #1;
        await Timer(1, "NS")
        # $write("%s\n", rw.convert2string());
#endclass

#class tb_env extends uvm_env;
class tb_env(UVMEnv):
    #
    def __init__(self, name="tb_env", parent=None):
        UVMEnv.__init__(self, name, parent)
        #   block_B   regmodel;
        #   reg_agent#(dut_rw) bus;
        #   uvm_reg_predictor#(reg_rw) predict;

    def build_phase(self, phase):
        self.regmodel = block_B.type_id.create("regmodel")
        self.regmodel.build()
        self.regmodel.lock_model()
        
        self.bus = reg_agent.type_id.create("bus", self)
        self.predict = UVMRegPredictor.type_id.create("predict", self)
        
        self.regmodel.set_hdl_path_root("dut");
        #  endfunction: build_phase

    def connect_phase(self, phase):
        reg2rw  = reg2rw_adapter("reg2rw")
        self.regmodel.default_map.set_sequencer(self.bus.sqr, reg2rw)
        
        self.predict.map = self.regmodel.default_map
        self.predict.adapter = reg2rw
        self.bus.mon.ap.connect(self.predict.bus_in)
        self.regmodel.default_map.set_auto_predict(0)

    #endclass
uvm_component_utils(tb_env)



#//----------------------------------------------------------------------
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2010 Mentor Graphics Corporation
#//   Copyright 2010-2011 Cadence Design Systems, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------


import cocotb
from uvm.comps.uvm_test import UVMTest
from uvm.macros import *
from uvm.base.uvm_root import uvm_top
from uvm.base.sv import sv
from sys_seqlib import *
from sys_env import *


class sys_R_test(UVMTest):

    def __init__(self, name="sys_R_test", parent=None):
        super().__init__(name, parent)
        self.env = None  # sys_env

    def build_phase(self, phase):
        env = []
        if self.env is None:
            if sv.cast(env, uvm_top.find("env$"), sys_env) is True:
                self.env = env[0]
            else:
                uvm_fatal("SYS_TEST/NO_ENV", "Unable to find/cast sys_env")

    
    async def run_phase(self, phase):
        # reset_seq = None  # uvm_sequence_base
        seq = None  # sys_R_test_seq
        phase.raise_objection(self)

        # dut_reset_seq rst_seq
        rst_seq = dut_reset_seq.type_id.create("rst_seq", self)
        rst_seq.vif = self.env.vif
        await rst_seq.start(None)
        self.env.model.reset()

        seq = sys_R_test_seq.type_id.create("sys_R_test_seq",self)
        seq.model = self.env.model
        await seq.start(None)

        phase.drop_objection(self)

    #endclass

uvm_component_utils(sys_R_test)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//


from uvm.comps.uvm_env import UVMEnv
from uvm.base.uvm_config_db import UVMConfigDb
from uvm.base.sv import sv
from uvm.macros import *

from blk_env import blk_env
from reg_S import *

from apb.apb_agent import apb_agent
from apb.apb_rw import reg2apb_adapter


class sys_env(UVMEnv):

    #   reg_sys_S model

    def __init__(self, name="sys_env", parent=None):
        super().__init__(name, parent)
        self.blk0 = None  # blk_env
        self.blk1 = None  # blk_env
        self.model = None  # reg_sys_S
        self.apb = None  # apb_agent
        self.vif = None
        self.all_ok = False


    def build_phase(self, phase):
        super().build_phase(phase)
        arr = []
        if UVMConfigDb.get(self, "apb", "vif", arr) is False:
            uvm_fatal("APB/SYS_ENV/NO_VIF", "Could not get vif from config_db")
        self.vif = arr[0]

        self.blk0 = blk_env.type_id.create("blk0", self)
        self.blk1 = blk_env.type_id.create("blk1", self)
        UVMConfigDb.set(self.blk0, "apb", "vif", arr[0])
        UVMConfigDb.set(self.blk1, "apb", "vif", arr[0])

        if self.model is None:
            self.model = reg_sys_S.type_id.create("reg_sys_S", None,
                self.get_full_name())
            self.model.build()
            self.model.lock_model()


        self.blk0.model = self.model.B[0]
        self.blk1.model = self.model.B[1]
        #
        self.apb = apb_agent.type_id.create("apb", self)
        #   endfunction: build_phase


    def connect_phase(self, phase):
        if self.model.get_parent() is None:
            reg2apb = reg2apb_adapter()
            self.model.default_map.set_sequencer(self.apb.sqr,reg2apb)
            self.model.default_map.set_auto_predict()

    def extract_phase(self, phase):
        if self.apb.mon.num_items > 0 and self.apb.mon.errors == 0:
            self.all_ok = True
        else:
            uvm_error("SYS_ENV_ERR", sv.sformatf("num_items: %0d, mon.errors: %0d",
                self.apb.mon.num_items, self.apb.mon.errors))

    #endclass: sys_env

uvm_component_utils(sys_env)



#//
#//----------------------------------------------------------------------------
#//   Copyright 2011 Mentor Graphics Corporation
#//   Copyright 2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------------

from uvm.reg.uvm_reg import UVMReg
from uvm.macros import uvm_object_utils
from uvm.base.sv import sv
from uvm.reg.uvm_reg_indirect import *
from uvm.reg.uvm_reg_field import UVMRegField
from uvm.reg.uvm_reg_file import *
from uvm.reg.uvm_mem import *
from uvm.reg.uvm_reg_block import *


class reg_slave_ID(UVMReg):


    def __init__(self, name="slave_ID"):
        super().__init__(name,32, UVM_NO_COVERAGE)
        self.REVISION_ID = None
        self.CHIP_ID = None
        self.PRODUCT_ID = None


    def build(self):
        self.REVISION_ID = UVMRegField.type_id.create("REVISION_ID")
        self.CHIP_ID = UVMRegField.type_id.create("CHIP_ID")
        self.PRODUCT_ID = UVMRegField.type_id.create("PRODUCT_ID")
        self.REVISION_ID.configure(self, 8, 0, "RO", 0, 0x03, 1, 0, 1)
        self.CHIP_ID.configure(self, 8, 8, "RO", 0, 0x5A, 1, 0, 1)
        self.PRODUCT_ID.configure(self, 10, 16,"RO", 0, 0x176, 1, 0, 1)


uvm_object_utils(reg_slave_ID)


class reg_slave_INDEX(UVMReg):


    def __init__(self, name="slave_INDEX"):
        super().__init__(name,8,UVM_NO_COVERAGE)


    def build(self):
        self.value = UVMRegField.type_id.create("value")
        self.value.configure(self, 8, 0, "RW", 0, 0x0, 1, 0, 1)


uvm_object_utils(reg_slave_INDEX)


class reg_slave_DATA(UVMRegIndirectData):

    def __init__(self, name="slave_DATA"):
        super().__init__(name,32, UVM_NO_COVERAGE)


uvm_object_utils(reg_slave_DATA)



class reg_slave_SOCKET(UVMReg):

    # TODO reg randomisation on sv side
    #   rand UVMRegField IP
    #   rand UVMRegField PORT

    def __init__(self, name="slave_ADDR"):
        super().__init__(name,64, UVM_NO_COVERAGE)
        self.IP = None
        self.PORT = None


    def build(self):
        self.IP   = UVMRegField.type_id.create("IP")
        self.PORT = UVMRegField.type_id.create("PORT")
        self.IP.configure(self, 48, 0, "RW", 0, 0x0, 1, 0, 1)
        self.PORT.configure(self, 16, 48, "RW", 0, 0x0, 1, 0, 1)


uvm_object_utils(reg_slave_SOCKET)


class reg_slave_SESSION(UVMRegFile):

    #   rand reg_slave_SOCKET SRC
    #   rand reg_slave_SOCKET DST

    def __init__(self, name="slave_SESSION"):
        super().__init__(name)


    def build(self):
        self.SRC = reg_slave_SOCKET.type_id.create("SRC")
        self.DST = reg_slave_SOCKET.type_id.create("DST")

        self.SRC.configure(self.get_block(), self, "SRC")
        self.DST.configure(self.get_block(), self, "DST")

        self.SRC.build()
        self.DST.build()


    def map(self, mp, offset):
        mp.add_reg(self.SRC, offset + 0x00)
        mp.add_reg(self.DST, offset + 0x08)


    def set_offset(self, mp, offset):
        self.SRC.set_offset(mp, offset + 0x00)
        self.DST.set_offset(mp, offset + 0x08)


uvm_object_utils(reg_slave_SESSION)


class reg_slave_TABLES(UVMReg):
    #
    #   rand UVMRegField value
    #
    def __init__(self, name="slave_TABLES"):
        super().__init__(name,32, UVM_NO_COVERAGE)
        self.value = None


    def build(self):
        self.value = UVMRegField.type_id.create("value")
        self.value.configure(self, 32, 0, "RW", 0, 0x0, 1, 0, 1)


uvm_object_utils(reg_slave_TABLES)


class mem_slave_DMA_RAM(UVMMem):
    #
    def __init__(self, name="slave_DMA_RAM"):
        super().__init__(name, 0x400, 32,"RW", UVM_NO_COVERAGE)


uvm_object_utils(mem_slave_DMA_RAM)


class reg_block_slave(UVMRegBlock):

    #   reg_slave_ID     ID
    #   reg_slave_INDEX  INDEX
    #   reg_slave_DATA   DATA

    #   rand reg_slave_SESSION  SESSION[256]

    #   rand reg_slave_TABLES   TABLES[256]

    #   mem_slave_DMA_RAM  DMA_RAM

    #   UVMRegField REVISION_ID
    #   UVMRegField CHIP_ID
    #   UVMRegField PRODUCT_ID

    def __init__(self, name="slave"):
        super().__init__(name, UVM_NO_COVERAGE)

        self.ID = None  # type: reg_slave_ID
        self.INDEX = None  # type: reg_slave_INDEX
        self.DATA = None  # type: reg_slave_DATA
        self.SESSION = []  # type: reg_slave_SESSION
        self.TABLES = []  # type: reg_slave_TABLES
        self.DMA_RAM = None  # type: mem_slave_DMA_RAM
        self.REVISION_ID = None  # type: UVMRegField
        self.CHIP_ID = None  # type: UVMRegField
        self.PRODUCT_ID = None  # type: UVMRegField
        self.nsession = 256


    def build(self):
        # create
        self.ID        = reg_slave_ID.type_id.create("ID")
        self.INDEX     = reg_slave_INDEX.type_id.create("INDEX")
        self.DATA      = reg_slave_DATA.type_id.create("DATA")
        for i in range(self.nsession):
            rslave_sess = reg_slave_SESSION.type_id.create(sv.sformatf("SESSION[%0d]",i))
            self.SESSION.append(rslave_sess)
            rslave_tables = reg_slave_TABLES.type_id.create(sv.sformatf("TABLES[%0d]",i))
            self.TABLES.append(rslave_tables)

        self.DMA_RAM   = mem_slave_DMA_RAM.type_id.create("DMA_RAM")
        #UVMResourceDb.set("REG::" + self.DMA_RAM.get_full_name() + ".*", "NO_REG_TESTS", 1, self)

        # configure
        self.ID.configure(self,None,"ID")
        self.ID.build()
        self.INDEX.configure(self,None,"INDEX")
        self.INDEX.build()
        for i in range(len(self.SESSION)):
            self.SESSION[i].configure(self,None,sv.sformatf("SESSION[%0d]",i))
            self.SESSION[i].build()

        for i in range(len(self.TABLES)):
            self.TABLES[i].configure(self, None, sv.sformatf("TABLES[%0d]",i))
            self.TABLES[i].build()

        self.DATA.configure(self.INDEX, self.TABLES, self, None)
        self.DATA.build()
        self.DMA_RAM.configure(self,"")

        # define default map
        self.default_map = self.create_map("default_map", 0x0, 4, UVM_LITTLE_ENDIAN)
        self.default_map.add_reg(self.ID, 0x0, "RW")
        self.default_map.add_reg(self.INDEX, 0x20, "RW")
        self.default_map.add_reg(self.DATA, 0x24, "RW")
        for i in range(len(self.SESSION)):
            self.SESSION[i].map(self.default_map, 0x1000 + 16 * i)

        self.default_map.add_mem(self.DMA_RAM, 0x2000, "RW")

        # field handle aliases
        self.REVISION_ID = self.ID.REVISION_ID
        self.CHIP_ID     = self.ID.CHIP_ID
        self.PRODUCT_ID  = self.ID.PRODUCT_ID


uvm_object_utils(reg_block_slave)



# ------------------------------------------------------------------------------
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#   uvm-python NOTE: All code ported from SystemVerilog UVM 1.2 to
#   python. Original code structures (including comments)
#   preserved where possible.
# ------------------------------------------------------------------------------


class UVMScopeStack:
    def __init__(self):
        self.m_arg = ""
        self.m_stack = []

    def depth(self):
        return len(self.m_stack)

    def get(self):
        v = ""
        if len(self.m_stack) == 0:
            return self.m_arg
        get = self.m_stack[0]

        for i in range(1, len(self.m_stack)):
            v = self.m_stack[i]
            if v != "" and (v[0] == "[" or v[0] == "(" or v[0] == "{"):
                get = get + v
            else:
                get = get + "." + v

        if self.m_arg != "":
            if get != "":
                get = get + "." + self.m_arg
            else:
                get = self.m_arg
        return get

    def get_arg(self):
        return self.m_arg

    def set(self, s):
        self.m_stack = []
        self.m_stack.append(s)
        self.m_arg = ""

    def down(self, s):
        self.m_stack.append(s)
        self.m_arg = ""

    def down_element(self, element):
        _str = "[{}]".format(element)
        self.m_stack.append(_str)
        self.m_arg = ""

    def up_element(self):
        s = ""
        if len(self.m_stack) == 0:
            return
        s = self.m_stack.pop()
        if s != "" and s[0] != "[":
            self.m_stack.append(s)

    def up(self, separator="."):
        found = False
        s = ""
        while len(self.m_stack) > 0 and found is False:
            s = self. m_stack.pop()
            if separator == ".":
                if s == "" or (s[0] != "[" and s[0] != "(" and s[0] != "{"):
                    found = True
            else:
                if s != "" and s[0] == separator:
                    found = True
        self.m_arg = ""


    def set_arg(self, arg):
        if arg == "" or arg is None:
            return
        self.m_arg = arg

    def set_arg_element(self, arg, ele):
        tmp_value_str = str(ele)
        self.m_arg = {arg, "[", tmp_value_str, "]"}

    def unset_arg(self, arg):
        if arg == self.m_arg:
            self.m_arg = ""



#//
#// -------------------------------------------------------------
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from .uvm_reg import UVMReg
from .uvm_reg_field import UVMRegField
from .uvm_reg_model import (UVM_CHECK, UVM_DEFAULT_PATH, UVM_NOT_OK,
    UVM_PREDICT_WRITE, UVM_PREDICT_DIRECT, UVM_PREDICT_READ)
from ..macros import uvm_warning, uvm_error
from ..base.sv import sv



class UVMRegFIFO(UVMReg):
    """
    Class: UVMRegFIFO

    This special register models a DUT FIFO accessed via write/read,
    where writes push to the FIFO and reads pop from it.

    Backdoor access is not enabled, as it is not yet possible to force
    complete FIFO state, i.e. the write and read indexes used to access
    the FIFO data.
    """



    #    // Variable: fifo
    #    //
    #    // The abstract representation of the FIFO. Constrained
    #    // to be no larger than the size parameter. It is public
    #    // to enable subtypes to add constraints on it and randomize.
    #    //
    #    rand uvm_reg_data_t fifo[$]


    #    constraint valid_fifo_size {
    #      len(self.fifo) <= m_size
    #    }


    #    //----------------------
    #    // Group: Initialization
    #    //----------------------

    #    // Function: new
    #    //
    #    // Creates an instance of a FIFO register having ~size~ elements of
    #    // ~n_bits~ each.
    #    //
    def __init__(self, name, size, n_bits, has_cover):
        super().__init__(name,n_bits,has_cover)
        self.m_size = size
        self.fifo = []
        self.m_set_cnt = 0


    #    // Funtion: build
    #    //
    #    // Builds the abstract FIFO register object. Called by
    #    // the instantiating block, a `UVMRegBlock` subtype.
    #    //
    def build(self):
        self.value = UVMRegField.type_id.create("value")
        self.value.configure(self, self.get_n_bits(), 0, "RW", 0, 0x0, 1, 0, 1)


    #    // Function: set_compare
    #    //
    #    // Sets the compare policy during a mirror (read) of the DUT FIFO.
    #    // The DUT read value is checked against its mirror only when both the
    #    // ~check~ argument in the <mirror()> call and the compare policy
    #    // for the field is <UVM_CHECK>.
    #    //
    def set_compare(self, check=UVM_CHECK):
        self.value.set_compare(check)


    #    //---------------------
    #    // Group: Introspection
    #    //---------------------

    #    // Function: size
    #    //
    #    // The number of entries currently in the FIFO.
    #    //
    def size(self):
        return len(self.fifo)


    #    // Function: capacity
    #    //
    #    // The maximum number of entries, or depth, of the FIFO.
    #
    def capacity(self):
        return self.m_size


    #    //--------------
    #    // Group: Access
    #    //--------------

    #    //  Function: write
    #    //
    #    //  Pushes the given value to the DUT FIFO. If auto-prediction is enabled,
    #    //  the written value is also pushed to the abstract FIFO before the
    #    //  call returns. If auto-prediction is not enabled (via
    #    //  `UVMRegMap.set_auto_predict`), the value is pushed to abstract
    #    //  FIFO only when the write operation is observed on the target bus.
    #    //  This mode requires using the <uvm_reg_predictor> class.
    #    //  If the write is via an `UVMRegFIFO.update` operation, the abstract FIFO
    #    //  already contains the written value and is thus not affected by
    #    //  either prediction mode.


    #    //  Function: read
    #    //
    #    //  Reads the next value out of the DUT FIFO. If auto-prediction is
    #    //  enabled, the frontmost value in abstract FIFO is popped.


    #    // Function: set
    #    //
    #    // Pushes the given value to the abstract FIFO. You may call this
    #    // method several times before an <update()> as a means of preloading
    #    // the DUT FIFO. Calls to ~set()~ to a full FIFO are ignored. You
    #    // must call <update()> to update the DUT FIFO with your set values.
    #    //
    def set(self, value, fname="", lineno=0):
        # emulate write, with intention of update
        value &= ((1 << self.get_n_bits())-1)
        if len(self.fifo) == self.m_size:
            return
        super().set(value,fname,lineno)
        self.m_set_cnt += 1
        self.fifo.append(self.value.value)


    #    // Function: update
    #    //
    #    // Pushes (writes) all values preloaded using <set()> to the DUT.
    #    // You must ~update~ after ~set~ before any blocking statements,
    #    // else other reads/writes to the DUT FIFO may cause the mirror to
    #    // become out of sync with the DUT.
    #    //
    async def update(self, status, path=UVM_DEFAULT_PATH, map=None, parent=None,
            prior=-1, extension=None, fname="", lineno=0):
        upd = 0x0
        if self.m_set_cnt == 0 or len(self.fifo) == 0:
            return
        self.m_update_in_progress = 1
        i = len(self.fifo)-self.m_set_cnt
        #for (int i=fifo.size()-m_set_cnt; m_set_cnt > 0; i++, m_set_cnt--):
        while self.m_set_cnt > 0:
            # uvm_reg_data_t val = get();
            # super.update(status,path,map,parent,prior,extension,fname,lineno);
            if i >= 0:
                status.clear()
                await self.write(status, self.fifo[i], path, map, parent,prior,extension,fname,lineno)
            i += 1
            self.m_set_cnt -= 1
        self.m_update_in_progress = 0


    #    // Function: mirror
    #    //
    #    // Reads the next value out of the DUT FIFO. If auto-prediction is
    #    // enabled, the frontmost value in abstract FIFO is popped. If
    #    // the ~check~ argument is set and comparison is enabled with
    #    // <set_compare()>.


    #    // Function: get
    #    //
    #    // Returns the next value from the abstract FIFO, but does not pop it.
    #    // Used to get the expected value in a <mirror()> operation.
    #    //
    def get(self, fname="", lineno=0):
        #return fifo.pop(0)();
        val = self.fifo[0]
        return val


    #    // Function: do_predict
    #    //
    #    // Updates the abstract (mirror) FIFO based on <write()> and
    #    // <read()> operations.  When auto-prediction is on, this method
    #    // is called before each read, write, peek, or poke operation returns.
    #    // When auto-prediction is off, this method is called by a
    #    // `UVMRegPredictor` upon receipt and conversion of an observed bus
    #    // operation to this register.
    #    //
    #    // If a write prediction, the observed
    #    // write value is pushed to the abstract FIFO as long as it is
    #    // not full and the operation did not originate from an <update()>.
    #    // If a read prediction, the observed read value is compared
    #    // with the frontmost value in the abstract FIFO if <set_compare()>
    #    // enabled comparison and the FIFO is not empty.
    #    //
    #    virtual function void do_predict(uvm_reg_item      rw,
    #                                     uvm_predict_e     kind = UVM_PREDICT_DIRECT,
    #                                     uvm_reg_byte_en_t be = -1)
    def do_predict(self, rw, kind=UVM_PREDICT_DIRECT, be=-1):
        super().do_predict(rw,kind,be)

        if rw.status == UVM_NOT_OK:
            return

        if kind in [UVM_PREDICT_WRITE, UVM_PREDICT_DIRECT]:
            if (len(self.fifo) != self.m_size) and (not self.m_update_in_progress):
                self.fifo.append(self.value.value)
        elif kind == UVM_PREDICT_READ:
            value = rw.value[0] & ((1 << self.get_n_bits())-1)
            mirror_val = 0x0
            if (len(self.fifo) == 0):
                return
            mirror_val = self.fifo.pop(0)
            if (self.value.get_compare() == UVM_CHECK and mirror_val != value):
                uvm_warning("MIRROR_MISMATCH", sv.sformatf(
                    "Observed DUT read value 'h%0h != mirror value 'h%0h, FIFO: %s",
                    value, mirror_val, str(self.fifo)))


    #    // Group: Special Overrides

    #    // Task: pre_write
    #    //
    #    // Special pre-processing for a <write()> or <update()>.
    #    // Called as a result of a <write()> or <update()>. It is an error to
    #    // attempt a write to a full FIFO or a write while an update is still
    #    // pending. An update is pending after one or more calls to <set()>.
    #    // If in your application the DUT allows writes to a full FIFO, you
    #    // must override `UVMRegFIFO.pre_write` as appropriate.
    async def pre_write(self, rw):
        if self.m_set_cnt > 0 and not self.m_update_in_progress:
            uvm_error("Needs Update","Must call update() after set() and before write()")
            rw.status = UVM_NOT_OK
            return

        if (len(self.fifo) >= self.m_size and not self.m_update_in_progress):
            uvm_error("FIFO Full","Write to full FIFO ignored")
            rw.status = UVM_NOT_OK
            return


    #    // Task: pre_read
    #    //
    #    // Special post-processing for a <write()> or <update()>.
    #    // Aborts the operation if the internal FIFO is empty. If in your application
    #    // the DUT does not behave this way, you must override `UVMRegFIFO.pre_read` as
    #    // appropriate.
    async def pre_read(self, rw):
        # abort if fifo empty
        if len(self.fifo) == 0:
            rw.status = UVM_NOT_OK
            return


    def post_randomize(self):
        self.m_set_cnt = 0

